from typing import Callable

import webviz_core_components as wcc
from dash import html

from ..models import PropertyStatisticsModel
from .selector_view import (
    ensemble_selector,
    filter_selector,
    property_selector,
    source_selector,
)


def selector_view(
    get_uuid: Callable, property_model: PropertyStatisticsModel
) -> html.Div:
    return [
        wcc.Selectors(
            label="Selectors",
            children=[
                ensemble_selector(
                    get_uuid=get_uuid,
                    ensembles=property_model.ensembles,
                    tab="qc",
                    multi=True,
                ),
                source_selector(
                    get_uuid=get_uuid, sources=property_model.sources, tab="qc"
                ),
                property_selector(
                    get_uuid=get_uuid, properties=property_model.properties, tab="qc"
                ),
            ],
        ),
        wcc.Selectors(
            label="Plot options",
            children=wcc.Checklist(
                id=get_uuid("property-qc-axis-match"),
                options=[{"label": "Shared plot axis", "value": "shared_axis"}],
                value=[],
            ),
        ),
        wcc.Selectors(
            label="Filters",
            children=filter_selector(
                get_uuid=get_uuid,
                property_model=property_model,
                tab="qc",
                value="Total" if property_model.selectors_has_value("Total") else None,
            ),
        ),
    ]


def property_qc_view(
    get_uuid: Callable, property_model: PropertyStatisticsModel
) -> wcc.FlexBox:
    return wcc.FlexBox(
        style={"margin": "20px"},
        children=[
            wcc.FlexColumn(
                children=wcc.Frame(
                    style={"height": "80vh", "overflowY": "auto"},
                    children=selector_view(
                        get_uuid=get_uuid, property_model=property_model
                    ),
                )
            ),
            wcc.FlexColumn(
                flex=8,
                children=wcc.Frame(
                    color="white",
                    highlight=False,
                    style={"height": "80vh"},
                    children=[
                        wcc.FlexBox(
                            children=[
                                wcc.RadioItems(
                                    vertical=False,
                                    id=get_uuid("property-qc-plot-type"),
                                    options=[
                                        {
                                            "label": "Distributions",
                                            "value": "distribution",
                                        },
                                        {
                                            "label": "Point per realization",
                                            "value": "scatter",
                                        },
                                        {
                                            "label": "Point per realization (single plot)",
                                            "value": "scatter_ensemble",
                                        },
                                        {"label": "Statistics table", "value": "table"},
                                    ],
                                    value="distribution",
                                    labelStyle={
                                        "display": "inline-block",
                                        "margin": "5px",
                                    },
                                ),
                            ]
                        ),
                        html.Div(
                            style={"height": "75vh"},
                            id=get_uuid("property-qc-wrapper"),
                        ),
                    ],
                ),
            ),
        ],
    )



from typing import Callable

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import webviz_core_components as wcc
from dash import Input, Output, State, callback, callback_context, html
from dash.exceptions import PreventUpdate

from ..utils.table_and_figure_utils import create_data_table, create_table_columns


def fipfile_qc_controller(get_uuid: Callable, disjoint_set_df: pd.DataFrame) -> None:
    @callback(
        Output(get_uuid("main-fipqc"), "children"),
        Input(get_uuid("selections"), "data"),
        Input({"id": get_uuid("main-fipqc"), "element": "display-option"}, "value"),
        State(get_uuid("page-selected"), "data"),
    )
    def _update_page_fipfileqc(
        selections: dict, display_option: str, page_selected: str
    ) -> html.Div:
        ctx = callback_context.triggered[0]

        if page_selected != "fipqc":
            raise PreventUpdate

        df = disjoint_set_df[["SET", "FIPNUM", "REGION", "ZONE", "REGZONE"]]

        selections = selections[page_selected]
        if not "display-option" in ctx["prop_id"]:
            if not selections["update"]:
                raise PreventUpdate

        for filt, values in selections["filters"].items():
            df = df.loc[df[filt].isin(values)]

        if selections["Group table"] and display_option == "table":
            df["FIPNUM"] = df["FIPNUM"].astype(str)
            df = df.groupby(["SET"]).agg(lambda x: ", ".join(set(x))).reset_index()

        df = df.sort_values(by=["SET"])

        if display_option == "table":
            return html.Div(
                children=create_data_table(
                    columns=create_table_columns(df.columns),
                    data=df.to_dict("records"),
                    height="82vh",
                    table_id={"table_id": "disjointset-info"},
                    style_cell_conditional=[
                        {"if": {"column_id": ["SET", "FIPNUM"]}, "width": "10%"},
                        {"if": {"column_id": ["ZONE", "REGION"]}, "width": "20%"},
                    ],
                    style_cell={
                        "whiteSpace": "normal",
                        "textAlign": "left",
                        "height": "auto",
                    },
                ),
            )

        df["FIPNUM"] = df["FIPNUM"].astype(str)
        return html.Div(
            [
                create_heatmap(df=df, y="ZONE", x="REGION"),
                create_heatmap(df=df, y="ZONE", x="FIPNUM"),
                create_heatmap(df=df, y="REGION", x="FIPNUM"),
            ]
        )


def create_heatmap(df: pd.DataFrame, y: str, x: str) -> wcc.Graph:
    """Create heatmap"""
    unique_y = df[y].unique()
    unique_x = sorted(df[x].unique(), key=int if x == "FIPNUM" else None)
    data = []
    for y_elm in unique_y:
        set_list = []
        for x_elm in unique_x:
            set_idx = df.loc[(df[y] == y_elm) & (df[x] == x_elm), "SET"]
            set_list.append(set_idx.iloc[0] if not set_idx.empty else None)
        data.append(set_list)

    return wcc.Graph(
        config={"displayModeBar": False},
        style={"height": "28vh"},
        figure=go.Figure(
            data=go.Heatmap(
                z=data,
                x=unique_x,
                y=unique_y,
                colorscale=(
                    px.colors.qualitative.Safe
                    + px.colors.qualitative.T10
                    + px.colors.qualitative.Set1
                ),
                showscale=False,
                hovertemplate="SET: %{z} <br>"
                + f"{x}: %{{x}} <br>"
                + f"{y}: %{{y}} <extra></extra>",
            )
        )
        .update_xaxes(title_text=x, tickangle=45, ticks="outside", **axis_variables())
        .update_yaxes(title_text=y, **axis_variables())
        .update_layout(
            margin={"l": 20, "r": 20, "t": 20, "b": 20}, plot_bgcolor="white"
        ),
    )


def axis_variables() -> dict:
    return {"showline": True, "linewidth": 2, "linecolor": "black", "mirror": True}



from webviz_config.utils import StrEnum


class PluginIds:  # pylint: disable=too-few-public-methods
    class Stores(StrEnum):
        SELECTED_ENSEMBLE = "selected-ensemble"
        SELECTED_WELLS = "selected-wells"
        SELECTED_MAX_NUMBER_OF_WELLS = "selected-max-number-of-wells"
        SELECTED_SORT_BY = "selected-sort-by"
        SELECTED_ASCENDING_DESCENDING = "selected-ascending-descending"
        SELECTED_STATISTICS = "selected-statistics"

    class SharedSettings(StrEnum):
        FILTER = "filter"
        VIEW_SETTINGS = "view-settings"

    class BhpId(StrEnum):
        LINE_CHART = "line-chart"
        FAN_CHART = "fan-chart"
        BAR_CHART = "bar-chart"



from typing import List, Union

import webviz_core_components as wcc
from dash import Input, Output, callback
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import ViewABC

from ..._reusable_settings import FilterLayout
from ..._reusable_view_element import GeneralViewElement
from ..._types import ColorAndSizeByType
from ..._utils import RftPlotterDataModel, filter_frame
from ._settings import Ensembles, PlotType, PlotTypeSettings, SizeColorSettings
from ._utils import update_crossplot, update_errorplot


class SimVsObsView(ViewABC):
    class Ids(StrEnum):
        PLOT_TYPE = "plot-type"
        ENSEMBLES = "ensembles"
        FILTERS = "filters"
        SIZE_COLOR_SETTINGS = "size-color-settings"
        VIEW_ELEMENT = "view-element"

    def __init__(self, datamodel: RftPlotterDataModel) -> None:
        super().__init__("Sim vs obs")
        self._datamodel = datamodel

        self.add_settings_groups(
            {
                self.Ids.PLOT_TYPE: PlotTypeSettings(),
                self.Ids.ENSEMBLES: Ensembles(self._datamodel.ensembles),
                self.Ids.FILTERS: FilterLayout(
                    wells=self._datamodel.well_names,
                    zones=self._datamodel.zone_names,
                    dates=self._datamodel.dates,
                ),
                self.Ids.SIZE_COLOR_SETTINGS: SizeColorSettings(),
            }
        )

        self.add_view_element(GeneralViewElement(), self.Ids.VIEW_ELEMENT)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element(self.Ids.VIEW_ELEMENT)
                .component_unique_id(GeneralViewElement.Ids.CHART)
                .to_string(),
                "children",
            ),
            Input(
                self.settings_group(self.Ids.PLOT_TYPE)
                .component_unique_id(PlotTypeSettings.Ids.PLOT_TYPE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.ENSEMBLES)
                .component_unique_id(Ensembles.Ids.ENSEMBLES)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.FILTERS)
                .component_unique_id(FilterLayout.Ids.FILTER_WELLS)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.FILTERS)
                .component_unique_id(FilterLayout.Ids.FILTER_ZONES)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.FILTERS)
                .component_unique_id(FilterLayout.Ids.FILTER_DATES)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.SIZE_COLOR_SETTINGS)
                .component_unique_id(SizeColorSettings.Ids.CROSSPLOT_SIZE_BY)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.SIZE_COLOR_SETTINGS)
                .component_unique_id(SizeColorSettings.Ids.CROSSPLOT_COLOR_BY)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _update_graph(
            plot_type: PlotType,
            ensembles: List[str],
            wells: List[str],
            zones: List[str],
            dates: List[str],
            sizeby: ColorAndSizeByType,
            colorby: ColorAndSizeByType,
        ) -> Union[str, List[wcc.Graph]]:
            df = filter_frame(
                self._datamodel.ertdatadf,
                {"WELL": wells, "ZONE": zones, "DATE": dates, "ENSEMBLE": ensembles},
            )
            if df.empty:
                return "No data matching the given filter criterias"

            if plot_type == PlotType.CROSSPLOT:
                return update_crossplot(df, sizeby, colorby)
            if plot_type == PlotType.ERROR_BOXPLOT:
                return [update_errorplot(df, self._datamodel.enscolors)]
            raise ValueError(f"Plot type: {plot_type.value} not implemented")



import copy
import datetime

import numpy as np
import pandas as pd
import pytest

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.dataframe_utils import make_date_column_datetime_object
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._types import (
    FanchartOptions,
    StatisticsOptions,
)

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.create_vector_traces_utils import (
    create_history_vector_trace,
    create_vector_fanchart_traces,
    create_vector_observation_traces,
    create_vector_realization_traces,
    create_vector_statistics_traces,
    render_hovertemplate,
)

# *******************************************************************
#####################################################################
#
# CONFIGURE TESTDATA
#
#####################################################################
# *******************************************************************

INPUT_VECTOR_STATISTICS_DF = pd.DataFrame(
    columns=[
        "DATE",
        StatisticsOptions.MEAN,
        StatisticsOptions.MIN,
        StatisticsOptions.MAX,
        StatisticsOptions.P10,
        StatisticsOptions.P90,
        StatisticsOptions.P50,
    ],
    data=[
        [datetime.datetime(2020, 1, 1), 1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        [datetime.datetime(2020, 1, 2), 1.5, 2.5, 3.5, 4.5, 5.5, 6.5],
    ],
)
make_date_column_datetime_object(INPUT_VECTOR_STATISTICS_DF)


# *******************************************************************
#####################################################################
#
# UNIT TESTS
#
#####################################################################
# *******************************************************************


def test_crate_vector_observation_traces() -> None:
    first_observation = {
        "date": datetime.datetime(2020, 1, 1),
        "value": 2.0,
        "comment": "first obs",
        "error": 0.5,
    }
    second_observation = {
        "date": datetime.datetime(2020, 6, 5),
        "value": 5.0,
        "comment": "second obs",
        "error": 1.2,
    }
    vector_observations = {"observations": [first_observation, second_observation]}

    expected_traces = [
        {
            "name": "Observation",
            "legendgroup": "Observation",
            "x": [datetime.datetime(2020, 1, 1), []],
            "y": [2.0, []],
            "marker": {"color": "black"},
            "hovertemplate": "(%{x}, %{y})<br>first obs",
            "showlegend": False,
            "error_y": {
                "type": "data",
                "array": [0.5, []],
                "visible": True,
            },
        },
        {
            "name": "Observation",
            "legendgroup": "Observation",
            "x": [datetime.datetime(2020, 6, 5), []],
            "y": [5.0, []],
            "marker": {"color": "black"},
            "hovertemplate": "(%{x}, %{y})<br>second obs",
            "showlegend": False,
            "error_y": {
                "type": "data",
                "array": [1.2, []],
                "visible": True,
            },
        },
    ]

    # Add legendgroup info for verification
    legend_group = "test group"
    expected_traces_with_legend_group = copy.deepcopy(expected_traces)
    expected_traces_with_legend_group[0]["name"] = "Observation: " + legend_group
    expected_traces_with_legend_group[0]["legendgroup"] = legend_group
    expected_traces_with_legend_group[1]["name"] = "Observation: " + legend_group
    expected_traces_with_legend_group[1]["legendgroup"] = legend_group

    assert expected_traces == create_vector_observation_traces(vector_observations)
    assert expected_traces_with_legend_group == create_vector_observation_traces(
        vector_observations, legend_group=legend_group
    )


def test_create_vector_realization_traces() -> None:
    vector_df = pd.DataFrame(
        columns=["DATE", "REAL", "A"],
        data=[
            [datetime.datetime(2020, 1, 1), 1, 1.0],
            [datetime.datetime(2020, 2, 1), 1, 2.0],
            [datetime.datetime(2031, 5, 10), 2, 5.0],
            [datetime.datetime(2031, 6, 10), 2, 6.0],
        ],
    )
    make_date_column_datetime_object(vector_df)

    created_traces = create_vector_realization_traces(
        vector_df=vector_df,
        ensemble="Test ensemble",
        color="red",
        legend_group="Test group",
        line_shape="linear",
        hovertemplate="Test hovertemplate ",
        show_legend=True,
        legendrank=2,
    )

    expected_traces = [
        {
            "line": {"width": 1, "shape": "linear", "color": "red"},
            "mode": "lines",
            "x": [datetime.datetime(2020, 1, 1), datetime.datetime(2020, 2, 1)],
            "y": [1.0, 2.0],
            "hovertemplate": "Test hovertemplate Realization: 1, Ensemble: Test ensemble",
            "name": "Test group",
            "legendgroup": "Test group",
            "legendrank": 2,
            "showlegend": True,
        },
        {
            "line": {"width": 1, "shape": "linear", "color": "red"},
            "mode": "lines",
            "x": [datetime.datetime(2031, 5, 10), datetime.datetime(2031, 6, 10)],
            "y": [5.0, 6.0],
            "hovertemplate": "Test hovertemplate Realization: 2, Ensemble: Test ensemble",
            "name": "Test group",
            "legendgroup": "Test group",
            "legendrank": 2,
            "showlegend": False,
        },
    ]

    assert expected_traces == created_traces


def test_create_vector_realization_traces_raise_error() -> None:
    multiple_vectors_df = pd.DataFrame(columns=["DATE", "REAL", "A", "B"])

    with pytest.raises(ValueError) as err:
        create_vector_realization_traces(
            vector_df=multiple_vectors_df,
            ensemble="Test ensemble",
            color="red",
            legend_group="Test group",
            line_shape="linear",
            hovertemplate="Test hovertemplate ",
        )
    assert str(err.value) == "Expected one vector column present in dataframe, got 2!"


def test_create_history_vector_trace() -> None:
    input_samples = [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 1, 2),
        datetime.datetime(2020, 1, 3),
    ]
    input_history_data = np.array([1.0, 2.0, 3.0])

    created_trace = create_history_vector_trace(
        samples=input_samples,
        history_data=input_history_data,
        line_shape="linear",
        color="green",
        vector_name="Test hist vector",
    )

    expected_trace = {
        "line": {"shape": "linear", "color": "green"},
        "mode": "lines",
        "x": input_samples,
        "y": input_history_data,
        "hovertext": "History: Test hist vector",
        "hoverinfo": "y+x+text",
        "name": "History",
        "showlegend": False,
        "legendgroup": "History",
        "legendrank": None,
    }

    assert created_trace == expected_trace


def test_create_history_vector_trace_raise_error() -> None:
    with pytest.raises(ValueError) as err:
        create_history_vector_trace(
            samples=[datetime.datetime(2020, 1, 1)],
            history_data=np.array([1.0, 2.0]),
            line_shape="linear",
            color="green",
            vector_name="Test hist vector",
        )
    assert str(err.value) == "Number of samples unequal number of data points!"


def test_render_hovertemplate() -> None:
    template_daily = "(%{x|%b} %{x|%-d}, %{x|%Y}, %{y})<br>"
    template_weekly = template_daily
    template_monthly = "(%{x|%b} %{x|%Y}, %{y})<br>"
    template_quarterly = "(Q%{x|%q} %{x|%Y}, %{y})<br>"
    template_yearly = "(%{x|%Y}, %{y})<br>"
    template_default = "(%{x}, %{y})<br>"

    # Test PER_DAY_/PER_INTVL_ vector names
    assert template_daily == render_hovertemplate("PER_DAY_a", Frequency.DAILY)
    assert template_weekly == render_hovertemplate("PER_DAY_a", Frequency.WEEKLY)
    assert template_monthly == render_hovertemplate("PER_DAY_a", Frequency.MONTHLY)
    assert template_quarterly == render_hovertemplate("PER_DAY_a", Frequency.QUARTERLY)
    assert template_yearly == render_hovertemplate("PER_DAY_a", Frequency.YEARLY)
    assert template_default == render_hovertemplate("PER_DAY_a", None)

    # Test other vector names
    assert template_default == render_hovertemplate("Vector_a", Frequency.DAILY)
    assert template_default == render_hovertemplate("Vector_a", Frequency.WEEKLY)
    assert template_default == render_hovertemplate("Vector_a", Frequency.MONTHLY)
    assert template_default == render_hovertemplate("Vector_a", Frequency.QUARTERLY)
    assert template_default == render_hovertemplate("Vector_a", Frequency.YEARLY)
    assert template_default == render_hovertemplate("Vector_a", None)


def test_create_vector_fanchart_traces_all_fanchart_options() -> None:
    created_fanchart_traces = create_vector_fanchart_traces(
        INPUT_VECTOR_STATISTICS_DF,
        fanchart_options=[
            FanchartOptions.MEAN,
            FanchartOptions.MIN_MAX,
            FanchartOptions.P10_P90,
        ],
        hex_color="#2b00ff",
        legend_group="First Legendgroup",
        line_shape="linear",
        show_legend=True,
    )

    # Verify number of trace dicts in list
    assert len(created_fanchart_traces) == 5

    # Verify Min
    assert created_fanchart_traces[0]["name"] == "First Legendgroup"
    assert created_fanchart_traces[0].get("fill", None) is None
    assert created_fanchart_traces[0]["hovertemplate"] == "(%{x}, %{y})<br>Minimum"
    assert np.array_equal(created_fanchart_traces[0]["y"], np.array([2.0, 2.5]))
    assert created_fanchart_traces[0]["x"] == [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 1, 2),
    ]

    # Verify P90
    assert created_fanchart_traces[1]["name"] == "First Legendgroup"
    assert created_fanchart_traces[1].get("fill", None) == "tonexty"
    assert created_fanchart_traces[1]["hovertemplate"] == "(%{x}, %{y})<br>P90"
    assert np.array_equal(created_fanchart_traces[1]["y"], np.array([5.0, 5.5]))
    assert created_fanchart_traces[1]["x"] == [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 1, 2),
    ]

    # Verify P10
    assert created_fanchart_traces[2]["name"] == "First Legendgroup"
    assert created_fanchart_traces[2].get("fill", None) == "tonexty"
    assert created_fanchart_traces[2]["hovertemplate"] == "(%{x}, %{y})<br>P10"
    assert np.array_equal(created_fanchart_traces[2]["y"], np.array([4.0, 4.5]))
    assert created_fanchart_traces[2]["x"] == [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 1, 2),
    ]

    # Verify Max
    assert created_fanchart_traces[3]["name"] == "First Legendgroup"
    assert created_fanchart_traces[3].get("fill", None) == "tonexty"
    assert created_fanchart_traces[3]["hovertemplate"] == "(%{x}, %{y})<br>Maximum"
    assert np.array_equal(created_fanchart_traces[3]["y"], np.array([3.0, 3.5]))
    assert created_fanchart_traces[3]["x"] == [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 1, 2),
    ]

    # Verify Mean
    assert created_fanchart_traces[4]["name"] == "First Legendgroup"
    assert created_fanchart_traces[4].get("fill", None) is None
    assert created_fanchart_traces[4]["hovertemplate"] == "(%{x}, %{y})<br>Mean"
    assert np.array_equal(created_fanchart_traces[4]["y"], np.array([1.0, 1.5]))
    assert created_fanchart_traces[4]["x"] == [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 1, 2),
    ]


def test_create_vector_fanchart_traces_subset_fanchart_options() -> None:
    created_fanchart_traces = create_vector_fanchart_traces(
        INPUT_VECTOR_STATISTICS_DF,
        fanchart_options=[
            FanchartOptions.MEAN,
            FanchartOptions.P10_P90,
        ],
        hex_color="#2b00ff",
        legend_group="Second Legendgroup",
        line_shape="linear",
        show_legend=True,
    )

    # Verify number of trace dicts in list
    assert len(created_fanchart_traces) == 3

    # Verify P90
    assert created_fanchart_traces[0]["name"] == "Second Legendgroup"
    assert created_fanchart_traces[0].get("fill", None) is None
    assert created_fanchart_traces[0]["hovertemplate"] == "(%{x}, %{y})<br>P90"
    assert np.array_equal(created_fanchart_traces[0]["y"], np.array([5.0, 5.5]))
    assert created_fanchart_traces[0]["x"] == [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 1, 2),
    ]

    # Verify P10
    assert created_fanchart_traces[1]["name"] == "Second Legendgroup"
    assert created_fanchart_traces[1].get("fill", None) == "tonexty"
    assert created_fanchart_traces[1]["hovertemplate"] == "(%{x}, %{y})<br>P10"
    assert np.array_equal(created_fanchart_traces[1]["y"], np.array([4.0, 4.5]))
    assert created_fanchart_traces[1]["x"] == [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 1, 2),
    ]

    # Verify Mean
    assert created_fanchart_traces[2]["name"] == "Second Legendgroup"
    assert created_fanchart_traces[2].get("fill", None) is None
    assert created_fanchart_traces[2]["hovertemplate"] == "(%{x}, %{y})<br>Mean"
    assert np.array_equal(created_fanchart_traces[2]["y"], np.array([1.0, 1.5]))
    assert created_fanchart_traces[2]["x"] == [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 1, 2),
    ]


def test_create_vector_statistics_traces_all_statistics_options() -> None:
    created_statistics_traces = create_vector_statistics_traces(
        INPUT_VECTOR_STATISTICS_DF,
        statistics_options=[
            StatisticsOptions.MEAN,
            StatisticsOptions.MIN,
            StatisticsOptions.MAX,
            StatisticsOptions.P10,
            StatisticsOptions.P50,
            StatisticsOptions.P90,
        ],
        color="green",
        legend_group="Third Legendgroup",
        line_shape="linear",
    )

    # Verify number of trace dicts in list
    assert len(created_statistics_traces) == 6

    # Verify Min
    assert created_statistics_traces[0]["name"] == "Third Legendgroup"
    assert created_statistics_traces[0]["hovertemplate"] == "(%{x}, %{y})<br>Minimum"
    assert np.array_equal(created_statistics_traces[0]["y"], np.array([2.0, 2.5]))
    assert np.array_equal(
        created_statistics_traces[0]["x"],
        np.array(
            [
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 2),
            ]
        ),
    )

    # Verify P90
    assert created_statistics_traces[1]["name"] == "Third Legendgroup"
    assert created_statistics_traces[1]["hovertemplate"] == "(%{x}, %{y})<br>P90"
    assert np.array_equal(created_statistics_traces[1]["y"], np.array([5.0, 5.5]))
    assert np.array_equal(
        created_statistics_traces[1]["x"],
        np.array(
            [
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 2),
            ]
        ),
    )

    # Verify P50
    assert created_statistics_traces[2]["name"] == "Third Legendgroup"
    assert created_statistics_traces[2]["hovertemplate"] == "(%{x}, %{y})<br>P50"
    assert np.array_equal(created_statistics_traces[2]["y"], np.array([6.0, 6.5]))
    assert np.array_equal(
        created_statistics_traces[2]["x"],
        np.array(
            [
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 2),
            ]
        ),
    )

    # Verify P10
    assert created_statistics_traces[3]["name"] == "Third Legendgroup"
    assert created_statistics_traces[3]["hovertemplate"] == "(%{x}, %{y})<br>P10"
    assert np.array_equal(created_statistics_traces[3]["y"], np.array([4.0, 4.5]))
    assert np.array_equal(
        created_statistics_traces[3]["x"],
        np.array(
            [
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 2),
            ]
        ),
    )

    # Verify Max
    assert created_statistics_traces[4]["name"] == "Third Legendgroup"
    assert created_statistics_traces[4]["hovertemplate"] == "(%{x}, %{y})<br>Maximum"
    assert np.array_equal(created_statistics_traces[4]["y"], np.array([3.0, 3.5]))
    assert np.array_equal(
        created_statistics_traces[4]["x"],
        np.array(
            [
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 2),
            ]
        ),
    )

    # Verify Mean
    assert created_statistics_traces[5]["name"] == "Third Legendgroup"
    assert created_statistics_traces[5]["hovertemplate"] == "(%{x}, %{y})<br>Mean"
    assert np.array_equal(created_statistics_traces[5]["y"], np.array([1.0, 1.5]))
    assert np.array_equal(
        created_statistics_traces[5]["x"],
        np.array(
            [
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 2),
            ]
        ),
    )


def test_create_vector_statistics_traces_subset_statistics_options() -> None:
    created_statistics_traces = create_vector_statistics_traces(
        INPUT_VECTOR_STATISTICS_DF,
        statistics_options=[
            StatisticsOptions.MEAN,
            StatisticsOptions.MIN,
            StatisticsOptions.P10,
        ],
        color="green",
        legend_group="Fourth Legendgroup",
        line_shape="linear",
    )

    # Verify number of trace dicts in list
    assert len(created_statistics_traces) == 3

    # Verify P50
    assert created_statistics_traces[0]["name"] == "Fourth Legendgroup"
    assert created_statistics_traces[0]["hovertemplate"] == "(%{x}, %{y})<br>Minimum"
    assert np.array_equal(created_statistics_traces[0]["y"], np.array([2.0, 2.5]))
    assert np.array_equal(
        created_statistics_traces[0]["x"],
        np.array(
            [
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 2),
            ]
        ),
    )

    # Verify P10
    assert created_statistics_traces[1]["name"] == "Fourth Legendgroup"
    assert created_statistics_traces[1]["hovertemplate"] == "(%{x}, %{y})<br>P10"
    assert np.array_equal(created_statistics_traces[1]["y"], np.array([4.0, 4.5]))
    assert np.array_equal(
        created_statistics_traces[1]["x"],
        np.array(
            [
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 2),
            ]
        ),
    )

    # Verify Mean
    assert created_statistics_traces[2]["name"] == "Fourth Legendgroup"
    assert created_statistics_traces[2]["hovertemplate"] == "(%{x}, %{y})<br>Mean"
    assert np.array_equal(created_statistics_traces[2]["y"], np.array([1.0, 1.5]))
    assert np.array_equal(
        created_statistics_traces[2]["x"],
        np.array(
            [
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 2),
            ]
        ),
    )



from webviz_config.utils import StrEnum


class PressureType(StrEnum):
    BHP = "BHP"
    DP = "DP"


class VfpParam(StrEnum):
    THP = "THP"
    WFR = "WFR"
    GFR = "GFR"
    ALQ = "ALQ"
    RATE = "RATE"



from typing import List, Optional


def printable_int_list(integer_list: Optional[List[int]]) -> str:
    """Creates a string out of a list of integers.
    The string gives a range x-y if all the integers between and
    including x and y are in the list, otherwise separated by commas.
    Example: the list [0, 1, 2, 4, 5, 8, 10] becomes '0-2, 4-5, 8, 10'
    If the input is `None` or the list is empty, the string 'None' is returned.
    """
    if not integer_list:
        return "None"

    sorted_list = sorted(integer_list)
    prev_number = sorted_list[0]
    string = str(prev_number)

    for next_number in sorted_list[1:]:
        if next_number > prev_number + 1:
            string += (
                f"{prev_number}" if string.endswith("-") else ""
            ) + f", {next_number}"
        elif not string.endswith("-"):
            string += "-"
        prev_number = next_number

    if not string.endswith(f", {prev_number}") and string != str(prev_number):
        string += str(prev_number)
    return string



import copy
from typing import Dict, List, Optional

import dash
import webviz_core_components as wcc
import webviz_subsurface_components as wsc
from dash import Input, Output, State, callback, dcc, html
from dash.development.base_component import Component
from dash.exceptions import PreventUpdate
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC
from webviz_subsurface_components import ExpressionInfo, ExternalParseData

from webviz_subsurface._utils.vector_calculator import (
    VectorDefinition,
    add_expressions_to_vector_selector_data,
    get_selected_expressions,
    get_vector_definitions_from_expressions,
)
from webviz_subsurface._utils.vector_selector import (
    is_vector_name_in_vector_selector_data,
)


def _create_new_selected_vectors(
    existing_selected_vectors: List[str],
    existing_expressions: List[ExpressionInfo],
    new_expressions: List[ExpressionInfo],
    new_vector_selector_data: list,
) -> List[str]:
    valid_selections: List[str] = []
    for vector in existing_selected_vectors:
        new_vector: Optional[str] = vector

        # Get id if vector is among existing expressions
        dropdown_id = next(
            (elm["id"] for elm in existing_expressions if elm["name"] == vector),
            None,
        )
        # Find id among new expressions to get new/edited name
        if dropdown_id:
            new_vector = next(
                (elm["name"] for elm in new_expressions if elm["id"] == dropdown_id),
                None,
            )

        # Append if vector name exist among data
        if new_vector is not None and is_vector_name_in_vector_selector_data(
            new_vector, new_vector_selector_data
        ):
            valid_selections.append(new_vector)
    return valid_selections


class TimeSeriesSettings(SettingsGroupABC):
    class Ids(StrEnum):
        VECTOR_SELECTOR = "vector-selector"
        VECTOR_CALCULATOR_OPEN_BUTTON = "vector-calculator-open-button"
        VECTOR_CALCULATOR_DIALOG = "vector-calculator-dialog"
        VECTOR_CALCULATOR = "vector-calculator"
        VECTOR_CALCULATOR_EXPRESSIONS_OPEN_DIALOG = (
            "vector-calculator-expression-open-dialog"
        )
        VECTOR_CALCULATOR_EXPRESSIONS = "vector-calculator-expression"
        GRAPH_DATA_HAS_CHANGED_TRIGGER = "graph-data-has-changed-trigger"

    def __init__(
        self,
        initial_vector_selector_data: List,
        custom_vector_definitions: Dict,
        vector_calculator_data: List,
        predefined_expressions: List[ExpressionInfo],
        vector_selector_base_data: List,
        custom_vector_definitions_base: Dict,
        initial_selected_vectors: Optional[List[str]] = None,
    ) -> None:
        super().__init__("Time series")
        self._vector_selector_data = initial_vector_selector_data
        self._selected_vectors = initial_selected_vectors
        self._custom_vector_definitions = custom_vector_definitions
        self._vector_calculator_data = vector_calculator_data
        self._predefined_expressions = predefined_expressions
        self._vector_selector_base_data = vector_selector_base_data
        self._custom_vector_definitions_base = custom_vector_definitions_base

    def layout(self) -> List[Component]:
        return [
            wsc.VectorSelector(
                id=self.register_component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_SELECTOR
                ),
                maxNumSelectedNodes=100,
                data=self._vector_selector_data,
                persistence=True,
                persistence_type="session",
                selectedTags=[]
                if self._selected_vectors is None
                else self._selected_vectors,
                numSecondsUntilSuggestionsAreShown=0.5,
                lineBreakAfterTag=True,
                customVectorDefinitions=self._custom_vector_definitions,
            ),
            html.Button(
                "Vector Calculator",
                id=self.register_component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_OPEN_BUTTON
                ),
                style={
                    "margin-top": "5px",
                    "margin-bottom": "5px",
                },
            ),
            wcc.Dialog(
                title="Vector Calculator",
                id=self.register_component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_DIALOG
                ),
                draggable=True,
                open=False,
                max_width="lg",
                children=[
                    html.Div(
                        style={"height": "60vh"},
                        children=[
                            wsc.VectorCalculator(
                                id=self.register_component_unique_id(
                                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR
                                ),
                                vectors=self._vector_calculator_data,
                                expressions=self._predefined_expressions,
                            )
                        ],
                    )
                ],
            ),
            dcc.Store(
                id=self.register_component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_EXPRESSIONS
                ),
                data=self._predefined_expressions,
            ),
            dcc.Store(
                id=self.register_component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_EXPRESSIONS_OPEN_DIALOG
                ),
                data=self._predefined_expressions,
            ),
            dcc.Store(
                # NOTE:Used to trigger graph update callback if data has
                # changed, i.e. no change of regular INPUT html-elements
                id=self.register_component_unique_id(
                    TimeSeriesSettings.Ids.GRAPH_DATA_HAS_CHANGED_TRIGGER
                ),
                data=0,
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_DIALOG
                ).to_string(),
                "open",
            ),
            Input(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_OPEN_BUTTON
                ).to_string(),
                "n_clicks",
            ),
            State(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_DIALOG
                ).to_string(),
                "open",
            ),
        )
        @callback_typecheck
        def _toggle_vector_calculator_dialog_open(
            n_open_clicks: Optional[int], is_open: bool
        ) -> bool:
            if n_open_clicks:
                return not is_open
            raise PreventUpdate

        @callback(
            Output(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR
                ).to_string(),
                "externalParseData",
            ),
            Input(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR
                ).to_string(),
                "externalParseExpression",
            ),
        )
        @callback_typecheck
        def _parse_vector_calculator_expression(
            expression: Optional[ExpressionInfo],
        ) -> ExternalParseData:
            if expression is None:
                raise PreventUpdate
            return wsc.VectorCalculator.external_parse_data(expression)

        @callback(
            Output(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_EXPRESSIONS
                ).to_string(),
                "data",
            ),
            Output(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_SELECTOR
                ).to_string(),
                "data",
            ),
            Output(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_SELECTOR
                ).to_string(),
                "selectedTags",
            ),
            Output(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_SELECTOR
                ).to_string(),
                "customVectorDefinitions",
            ),
            Output(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.GRAPH_DATA_HAS_CHANGED_TRIGGER
                ).to_string(),
                "data",
            ),
            Input(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_DIALOG
                ).to_string(),
                "open",
            ),
            State(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_EXPRESSIONS_OPEN_DIALOG
                ).to_string(),
                "data",
            ),
            State(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_EXPRESSIONS
                ).to_string(),
                "data",
            ),
            State(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_SELECTOR
                ).to_string(),
                "selectedNodes",
            ),
            State(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_SELECTOR
                ).to_string(),
                "customVectorDefinitions",
            ),
            State(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.GRAPH_DATA_HAS_CHANGED_TRIGGER
                ).to_string(),
                "data",
            ),
        )
        @callback_typecheck
        def _update_vector_calculator_expressions_on_dialog_close(
            is_dialog_open: bool,
            new_expressions: List[ExpressionInfo],
            current_expressions: List[ExpressionInfo],
            current_selected_vectors: List[str],
            current_custom_vector_definitions: Dict[str, VectorDefinition],
            graph_data_has_changed_counter: int,
        ) -> list:
            """Update vector calculator expressions, propagate expressions to VectorSelectors,
            update current selections and trigger re-rendering of graphing if necessary
            """

            if is_dialog_open or (new_expressions == current_expressions):
                raise PreventUpdate

            # Create current selected expressions for comparison - Deep copy!
            current_selected_expressions = get_selected_expressions(
                current_expressions, current_selected_vectors
            )

            # Create new vector selector data - Deep copy!
            new_vector_selector_data = copy.deepcopy(self._vector_selector_base_data)
            add_expressions_to_vector_selector_data(
                new_vector_selector_data, new_expressions
            )

            # Create new selected vectors - from new expressions
            new_selected_vectors = _create_new_selected_vectors(
                current_selected_vectors,
                current_expressions,
                new_expressions,
                new_vector_selector_data,
            )

            # Get new selected expressions
            new_selected_expressions = get_selected_expressions(
                new_expressions, new_selected_vectors
            )

            # Get new custom vector definitions
            new_custom_vector_definitions = get_vector_definitions_from_expressions(
                new_expressions
            )
            for key, value in self._custom_vector_definitions_base.items():
                if key not in new_custom_vector_definitions:
                    new_custom_vector_definitions[key] = value

            # Prevent updates if unchanged
            if new_custom_vector_definitions == current_custom_vector_definitions:
                new_custom_vector_definitions = dash.no_update

            if new_selected_vectors == current_selected_vectors:
                new_selected_vectors = dash.no_update

            # If selected expressions are edited
            # - Only trigger graph data update property when needed,
            # i.e. names are unchanged and selectedNodes for VectorSelector remains unchanged.
            new_graph_data_has_changed_counter = dash.no_update
            if (
                new_selected_expressions != current_selected_expressions
                and new_selected_vectors == dash.no_update
            ):
                new_graph_data_has_changed_counter = graph_data_has_changed_counter + 1

            return [
                new_expressions,
                new_vector_selector_data,
                new_selected_vectors,
                new_custom_vector_definitions,
                new_graph_data_has_changed_counter,
            ]

        @callback(
            Output(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_EXPRESSIONS_OPEN_DIALOG
                ).to_string(),
                "data",
            ),
            Input(
                self.component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR
                ).to_string(),
                "expressions",
            ),
        )
        @callback_typecheck
        def _update_vector_calculator_expressions_when_dialog_open(
            expressions: List[ExpressionInfo],
        ) -> list:
            new_expressions: List[ExpressionInfo] = [
                elm for elm in expressions if elm["isValid"]
            ]
            return new_expressions



import io
import logging

import numpy as np
import xtgeo
from PIL import Image

from webviz_subsurface._utils.perf_timer import PerfTimer

# !!!!!!!
# This is basically a copy of surface_to_rgba() from _ensemble_surface_plugin._make_rgba.py
# with a slight change in signature

LOGGER = logging.getLogger(__name__)


def surface_to_png_bytes(surface: xtgeo.RegularSurface) -> bytes:
    """Converts a xtgeo Surface to RGBA array. Used to set the image when used in a
    DeckGLMap component"""

    timer = PerfTimer()

    # surface.unrotate()
    LOGGER.debug(f"unrotate: {timer.lap_s():.2f}s")

    surface.fill(np.nan)
    values = surface.values
    values = np.flip(values.transpose(), axis=0)

    # If all values are masked set to zero
    if values.mask.all():
        values = np.zeros(values.shape)

    LOGGER.debug(f"fill/flip/mask: {timer.lap_s():.2f}s")

    min_val = np.nanmin(surface.values)
    max_val = np.nanmax(surface.values)
    if min_val == 0.0 and max_val == 0.0:
        scale_factor = 1.0
    else:
        scale_factor = (256 * 256 * 256 - 1) / (max_val - min_val)

    LOGGER.debug(f"minmax: {timer.lap_s():.2f}s")

    z_array = (values.copy() - min_val) * scale_factor
    z_array = z_array.copy()
    shape = z_array.shape

    LOGGER.debug(f"scale and copy: {timer.lap_s():.2f}s")

    z_array = np.repeat(z_array, 4)  # This will flatten the array

    z_array[0::4][np.isnan(z_array[0::4])] = 0  # Red
    z_array[1::4][np.isnan(z_array[1::4])] = 0  # Green
    z_array[2::4][np.isnan(z_array[2::4])] = 0  # Blue

    z_array[0::4] = np.floor((z_array[0::4] / (256 * 256)) % 256)  # Red
    z_array[1::4] = np.floor((z_array[1::4] / 256) % 256)  # Green
    z_array[2::4] = np.floor(z_array[2::4] % 256)  # Blue
    z_array[3::4] = np.where(np.isnan(z_array[3::4]), 0, 255)  # Alpha

    LOGGER.debug(f"bytestuff: {timer.lap_s():.2f}s")

    # Back to 2d shape + 1 dimension for the rgba values.

    z_array = z_array.reshape((shape[0], shape[1], 4))

    image = Image.fromarray(np.uint8(z_array), "RGBA")
    LOGGER.debug(f"create: {timer.lap_s():.2f}s")

    byte_io = io.BytesIO()
    # Huge speed benefit from reducing compression level
    image.save(byte_io, format="png", compress_level=1)
    LOGGER.debug(f"save png to bytes: {timer.lap_s():.2f}s")

    byte_io.seek(0)
    ret_bytes = byte_io.read()
    LOGGER.debug(f"read bytes: {timer.lap_s():.2f}s")

    LOGGER.debug(f"Total time: {timer.elapsed_s():.2f}s")

    return ret_bytes


# pylint: disable=too-many-locals
def surface_to_png_bytes_optimized(surface: xtgeo.RegularSurface) -> bytes:

    timer = PerfTimer()
    # Note that returned values array is a 2d masked array
    surf_values_ma: np.ma.MaskedArray = surface.values

    surf_values_ma = np.flip(surf_values_ma.transpose(), axis=0)  # type: ignore
    LOGGER.debug(f"flip/transpose: {timer.lap_s():.2f}s")

    # This will be a flat bool array with true for all valid entries
    valid_arr = np.invert(np.ma.getmaskarray(surf_values_ma).flatten())
    LOGGER.debug(f"get valid_arr: {timer.lap_s():.2f}s")

    shape = surf_values_ma.shape
    min_val = surf_values_ma.min()
    max_val = surf_values_ma.max()
    LOGGER.debug(f"minmax: {timer.lap_s():.2f}s")

    if min_val == 0.0 and max_val == 0.0:
        scale_factor = 1.0
    else:
        scale_factor = (256 * 256 * 256 - 1) / (max_val - min_val)

    # Scale the values into the wanted range
    scaled_values_ma = (surf_values_ma - min_val) * scale_factor

    # Get a NON-masked array with all undefined entries filled with 0
    scaled_values = scaled_values_ma.filled(0)

    LOGGER.debug(f"scale and fill: {timer.lap_s():.2f}s")

    val_arr = scaled_values.astype(np.uint32).ravel()
    LOGGER.debug(f"cast and flatten: {timer.lap_s():.2f}s")

    val = val_arr.view(dtype=np.uint8)
    rgba_arr = np.empty(4 * len(val_arr), dtype=np.uint8)
    rgba_arr[0::4] = val[2::4]
    rgba_arr[1::4] = val[1::4]
    rgba_arr[2::4] = val[0::4]
    rgba_arr[3::4] = np.multiply(valid_arr, 255).astype(np.uint8)

    LOGGER.debug(f"rgba combine: {timer.lap_s():.2f}s")

    # Back to 2d shape + 1 dimension for the rgba values.
    rgba_arr_reshaped = rgba_arr.reshape((shape[0], shape[1], 4))

    image = Image.fromarray(rgba_arr_reshaped, "RGBA")
    LOGGER.debug(f"create: {timer.lap_s():.2f}s")

    byte_io = io.BytesIO()
    image.save(byte_io, format="png", compress_level=1)
    LOGGER.debug(f"save png to bytes: {timer.lap_s():.2f}s")

    byte_io.seek(0)
    ret_bytes = byte_io.read()
    LOGGER.debug(f"read bytes: {timer.lap_s():.2f}s")

    LOGGER.debug(f"Total time: {timer.elapsed_s():.2f}s")

    return ret_bytes



# pylint: disable=too-many-lines
import io
import json
from pathlib import Path
from typing import Callable, List, Tuple
from uuid import uuid4

import dash_bootstrap_components as dbc
import defusedxml.ElementTree as ET
import numpy as np
import webviz_core_components as wcc
import webviz_subsurface_components
import xtgeo
from dash import Dash, Input, Output, State, callback_context, dash_table, dcc, html
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.deprecation_decorators import deprecated_plugin
from webviz_config.webviz_assets import WEBVIZ_ASSETS
from webviz_config.webviz_store import webvizstore

import webviz_subsurface
from webviz_subsurface._datainput.well import get_well_layers
from webviz_subsurface._models import SurfaceLeafletModel

from ._huv_table import FilterTable
from ._huv_xsection import HuvXsection


@deprecated_plugin(
    "Relevant functionality is implemented in the StructuralUncertainty plugin."
)
class HorizonUncertaintyViewer(WebvizPluginABC):
    """Visualizes depth uncertainty for surfaces in map view and cross section view.

    The cross section is defined by wellfiles and surfacefiles or a polyline.
    Polylines are drawn interactivly in map view.

    !> The plugin reads information from a COHIBA model file.

    * **`basedir`:** Path to folder with model_file.xml.
       Make sure that the folder has the same format as a COHIBA folder.
    * **`planned_wells_dir`:** Path to folder with planned well files.
       Make sure that all planned wells have format 'ROXAR RMS well'."""

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        basedir: Path,
        planned_wells_dir: Path = None,
    ):

        super().__init__()
        self.plotly_theme = webviz_settings.theme.plotly_theme
        self.uid = uuid4()
        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent / "_assets" / "css" / "modal.css"
        )
        self.set_callbacks(app)

        self.basedir = basedir
        self.planned_wells_dir = planned_wells_dir
        self.modelfile_path = basedir / "model_file.xml"
        self.modelfile = get_path(self.modelfile_path)
        self.surfaces = load_surfaces(basedir, self.modelfile_path)
        self.planned_wellfiles = (
            json.load(find_files(planned_wells_dir, "*.txt"))
            if planned_wells_dir
            else None
        )
        self.wellfiles = json.load(find_files(basedir / "input" / "welldata", "*.txt"))
        self.wellfiles = [str(get_path(Path(w))) for w in self.wellfiles]
        self.allfiles = json.load(find_files(basedir))
        self.allfiles.append(self.modelfile_path)
        self.allfiles += self.planned_wellfiles
        self.planned_wellfiles = [
            str(get_path(Path(w))) for w in self.planned_wellfiles
        ]
        self.surface_attributes = {}
        for i, surface in enumerate(self.surfaces):
            self.surface_attributes[surface["name"]] = {
                "color": get_color(i),
                "order": i,
                "name": surface["name"],
                "topofzone": surface["topofzone"],
                "surface": surface["d_"],
                "surface_de": surface["de_"],
                "surface_dt": surface["dt_"],
                "surface_dr": surface["dr_"],
                "surface_dte": surface["dte_"],
            }

        self.surfacenames = [surface["name"] for surface in self.surfaces]
        # Log files
        zonation_status_file = get_zonation_status(basedir)
        well_points_file = get_well_points(basedir)
        zonelog_name = get_zonelog_name(self.modelfile)
        self.xsec = HuvXsection(
            self.surface_attributes,
            zonation_status_file,
            well_points_file,
            zonelog_name,
        )
        target_points_file = get_target_points(basedir)
        self.df_well_target_points = FilterTable(target_points_file, well_points_file)

        # Wellfiles and planned wells
        self.planned_wells = {}
        if planned_wells_dir is not None:
            self.planned_wells = {
                wf: xtgeo.well_from_file(wfile=wf) for wf in self.planned_wellfiles
            }

        self.wells = {wf: xtgeo.well_from_file(wfile=wf) for wf in self.wellfiles}

        # Store current layers
        self.state = {"switch": False}
        self.layers_state = []

    def ids(self, element: str) -> str:
        return f"{element}-id-{self.uid}"

    @property
    def surface_types(self) -> List[str]:
        return [
            "Depth",
            "Depth uncertainty",
            "Depth residual",
            "Depth residual uncertainty",
            "Depth trend",
            "Depth trend uncertainty",
        ]

    @property
    def cross_section_graph_layout(self) -> html.Div:
        return html.Div(
            children=[
                wcc.Graph(
                    id=self.ids("xsec-view"),
                )
            ]
        )

    @property
    def cross_section_widgets_layout(self) -> html.Div:
        return html.Div(
            children=[
                html.Div(
                    children=[
                        dbc.Button(
                            "Surface Settings",
                            id=self.ids("button-open-graph-settings"),
                            color="light",
                            className="mr-1",
                        ),
                        dbc.Modal(
                            children=[
                                dbc.ModalHeader("Surface Settings"),
                                dbc.ModalBody(
                                    children=[
                                        html.Label(
                                            style={
                                                "font-weight": "bold",
                                                "textAlign": "Left",
                                            },
                                            children="Select Surfaces",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("all-surfaces-checkbox"),
                                            options=[{"label": "all", "value": "True"}],
                                            value=["True"],
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("surfaces-checklist"),
                                            options=[
                                                {"label": name, "value": name}
                                                for name in self.surfacenames
                                            ],
                                            value=self.surfacenames,
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("surfaces-de-checklist"),
                                            options=[
                                                {
                                                    "label": name + " SD",
                                                    "value": name,
                                                    "disabled": False,
                                                }
                                                for name in self.surfacenames
                                            ],
                                            value=self.surfacenames,
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                    ],
                                ),
                                dbc.ModalFooter(
                                    children=[
                                        dbc.Button(
                                            "Close",
                                            id=self.ids("button-close-graph-settings"),
                                            className="ml-auto",
                                        ),
                                        dbc.Button(
                                            "Apply changes",
                                            id=self.ids("button-apply-checklist"),
                                            className="ml-auto",
                                        ),
                                    ]
                                ),
                            ],
                            id=self.ids("modal-graph-settings"),
                            size="sm",
                            centered=True,
                            backdrop=False,
                            fade=False,
                        ),
                        dbc.Button(
                            "Well Settings", id=self.ids("button-open-well-settings")
                        ),
                        dbc.Modal(
                            children=[
                                dbc.ModalHeader("Well Settings"),
                                dbc.ModalBody(
                                    children=[
                                        html.Label(
                                            style={
                                                "font-weight": "bold",
                                                "textAlign": "Left",
                                            },
                                            children="Select Well Attributes",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("all-well-settings-checkbox"),
                                            options=[{"label": "all", "value": "True"}],
                                            value=["True"],
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("well-settings-checklist"),
                                            options=[
                                                {
                                                    "label": "Zonelog",
                                                    "value": "zonelog",
                                                },
                                                {
                                                    "label": "Zonation points",
                                                    "value": "zonation_points",
                                                },
                                                {
                                                    "label": "Conditional points",
                                                    "value": "conditional_points",
                                                },
                                            ],
                                            value=[
                                                "zonelog",
                                                "zonation_points",
                                                "conditional_points",
                                            ],
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                    ],
                                ),
                                dbc.ModalFooter(
                                    children=[
                                        dbc.Button(
                                            "Close",
                                            id=self.ids("button-close-well-settings"),
                                            className="ml-auto",
                                        ),
                                        dbc.Button(
                                            "Apply",
                                            id=self.ids(
                                                "button-apply-well-settings-checklist"
                                            ),
                                            className="ml-auto",
                                        ),
                                    ]
                                ),
                            ],
                            id=self.ids("modal-well-settings"),
                            size="sm",
                            centered=True,
                            backdrop=False,
                            fade=False,
                        ),
                    ],
                ),
                wcc.FlexBox(
                    children=[
                        html.Div(
                            children=[
                                html.Label(
                                    style={
                                        "font-weight": "bold",
                                        "textAlign": "center",
                                    },
                                    children="Select well",
                                ),
                                dcc.Dropdown(
                                    id=self.ids("well-dropdown"),
                                    options=[
                                        {
                                            "label": self.wells[wf].wellname,
                                            "value": str(wf),
                                        }
                                        for wf in self.wellfiles
                                    ]
                                    + [
                                        {
                                            "label": self.planned_wells[wf].wellname,
                                            "value": str(wf),
                                        }
                                        for wf in self.planned_wellfiles
                                    ],
                                    value=str(self.wellfiles[0]),
                                    clearable=False,
                                    disabled=False,
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ]
                        ),
                    ],
                ),
                html.Div(
                    children=[
                        html.Div(
                            style={
                                "marginTop": "0px",
                                "height": "800px",
                                "zIndex": -9999,
                            },
                            children=[self.cross_section_graph_layout],
                            id=self.ids("cross-section-view"),
                        )
                    ]
                ),
            ]
        )

    @property
    def target_points_tab_layout(self) -> dash_table.DataTable:
        df = self.df_well_target_points.get_targetpoints_df()
        return dash_table.DataTable(
            id=self.ids("target-point-table"),
            columns=[{"name": i, "id": i} for i in df.columns],
            data=df.to_dict("records"),
            sort_action="native",
            filter_action="native",
        )

    @property
    def well_points_tab_layout(self) -> html.Div:
        return html.Div(
            [
                dbc.Button("Table Settings", id=self.ids("button-open-table-settings")),
                dbc.Modal(
                    children=[
                        dbc.ModalHeader("Table Settings"),
                        dbc.ModalBody(
                            children=[
                                html.Label(
                                    style={
                                        "font-weight": "bold",
                                        "textAlign": "Left",
                                    },
                                    children="Select Table Columns",
                                ),
                                dcc.Checklist(
                                    id=self.ids("columns-checklist"),
                                    options=[
                                        {"label": name, "value": column_name}
                                        for name, column_name in zip(
                                            self.df_well_target_points.get_wellpoints_df()
                                            .keys()
                                            .values,
                                            self.df_well_target_points.get_wellpoints_df()
                                            .keys()
                                            .values,
                                        )
                                    ],
                                    value=[
                                        "Surface",
                                        "Well",
                                        "TVD",
                                        "MD",
                                        "Outlier",
                                        "Deleted",
                                        "Residual",
                                    ],
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ],
                        ),
                        dbc.ModalFooter(
                            children=[
                                dbc.Button(
                                    "Close",
                                    id=self.ids("button-close-table-settings"),
                                    className="ml-auto",
                                ),
                                dbc.Button(
                                    "Apply",
                                    id=self.ids("button-apply-columnlist"),
                                    className="ml-auto",
                                ),
                            ]
                        ),
                    ],
                    id=self.ids("modal-table-settings"),
                    size="sm",
                    centered=True,
                    backdrop=False,
                    fade=False,
                ),
                html.Div(id=self.ids("well-points-table-container")),
            ]
        )

    @property
    def left_flexbox_layout(self) -> html.Div:
        return html.Div(
            children=[
                wcc.FlexBox(
                    children=[
                        dcc.RadioItems(
                            labelStyle={"display": "inline-block"},
                            options=[
                                {"label": "Map view", "value": "map-view"},
                                {"label": "Surface picks table", "value": "table-view"},
                            ],
                            id=self.ids("map-table-radioitems"),
                            value="map-view",
                            persistence=True,
                            persistence_type="session",
                        )
                    ],
                    style={"padding": "5px"},
                ),
                html.Div(
                    id=self.ids("hidden-div-map-view"), children=[self.map_view_layout]
                ),  # Hidden div to store polyline points when in table view
                html.Div(
                    id=self.ids("hidden-div-table-view"),
                    children=[self.table_view_layout],
                ),
            ]
        )

    @property
    def map_view_layout(self) -> html.Div:
        return html.Div(
            children=[
                wcc.FlexBox(
                    children=[
                        html.Div(
                            children=[
                                html.Label(
                                    style={
                                        "font-weight": "bold",
                                        "textAlign": "center",
                                    },
                                    children="Select surface",
                                ),
                                dcc.Dropdown(
                                    id=self.ids("map-dropdown"),
                                    options=[
                                        {"label": name, "value": name}
                                        for name in self.surfacenames
                                    ],
                                    value=self.surfacenames[0],
                                    clearable=False,
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ]
                        ),
                    ],
                ),
                html.Div(
                    style={
                        "marginTop": "0px",
                        "height": "800px",
                        "zIndex": -9999,
                    },
                    children=[
                        webviz_subsurface_components.LeafletMap(
                            id=self.ids("layered-map"),
                            layers=[],
                            unitScale={},
                            autoScaleMap=True,
                            minZoom=-19,
                            drawTools={
                                "drawMarker": False,
                                "drawPolygon": False,
                                "drawPolyline": True,
                                "position": "topright",
                            },
                            switch={
                                "value": False,
                                "disabled": False,
                                "label": "Hillshading",
                            },
                            mouseCoords={"position": "bottomright"},
                            colorBar={"position": "bottomright"},
                        ),
                    ],
                ),
            ]
        )

    @property
    def table_view_layout(self) -> html.Div:
        df = self.xsec.get_intersection_dataframe(self.wells[self.wellfiles[0]])
        return html.Div(
            children=[
                html.Label(
                    id=self.ids("surface-picks-label"),
                    style={
                        "font-weight": "bold",
                        "textAlign": "center",
                    },
                ),
                dash_table.DataTable(
                    id=self.ids("uncertainty-table"),
                    columns=[{"name": i, "id": i} for i in df.columns],
                    data=df.to_dict("records"),
                    sort_action="native",
                    filter_action="native",
                ),
            ],
            style={"marginTop": "0px"},
        )

    @property
    def layout(self) -> dcc.Tabs:
        return dcc.Tabs(
            children=[
                dcc.Tab(
                    label="Cross section & map view",
                    children=[
                        wcc.FlexBox(
                            id=self.ids("layout"),
                            children=[
                                html.Div(
                                    style={"flex": 1}, children=self.left_flexbox_layout
                                ),
                                html.Div(
                                    style={"flex": 1.5},
                                    children=self.cross_section_widgets_layout,
                                ),
                            ],
                        )
                    ],
                ),
                dcc.Tab(
                    label="Target Points",
                    children=[html.Div(children=self.target_points_tab_layout)],
                ),
                dcc.Tab(label="Well Points", children=[self.well_points_tab_layout]),
            ]
        )

    # pylint: disable=too-many-statements
    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.ids("layered-map"), "layers"),
            [
                Input(self.ids("map-dropdown"), "value"),
                Input(self.ids("layered-map"), "switch"),  # Toggle hillshading on/off
                Input(self.ids("well-dropdown"), "value"),  # Wellfile
            ],
        )
        def _render_map(surface_name, switch, wellfile):
            """Renders map view for one surface with de, dt, dte, dr, dre and depth
            Wells marked with circles, trajectory and hillshading toggle
            """

            surfaces = [
                self.surface_attributes[surface_name]["surface"],
                self.surface_attributes[surface_name]["surface_de"],
                self.surface_attributes[surface_name]["surface_dr"],
                self.surface_attributes[surface_name]["surface_dt"],
                self.surface_attributes[surface_name]["surface_dte"],
            ]
            well_layers = get_well_layers(
                self.wells,
                self.planned_wells,
                surface_name,
                surfaces[0],
                wellfile,
            )
            layers = []
            for i, surface in enumerate(surfaces):
                layers.append(
                    SurfaceLeafletModel(
                        surface,
                        name=self.surface_types[i],
                        apply_shading=switch["value"],
                        updatemode="add",
                    ).layer
                )
            layers.extend(well_layers)
            # Deletes old layers when switching surface in dropdown
            old_layers = self.layers_state
            self.layers_state = layers.copy()
            if old_layers is not None and len(old_layers) > 0:
                for layer in old_layers:
                    layer["action"] = "delete"
                old_layers.extend(layers)
                layers = old_layers
            return layers

        @app.callback(
            Output(self.ids("xsec-view"), "figure"),
            [
                Input(self.ids("button-apply-checklist"), "n_clicks"),
                Input(self.ids("button-apply-well-settings-checklist"), "n_clicks"),
                Input(self.ids("well-dropdown"), "value"),  # wellpath
                Input(self.ids("layered-map"), "polyline_points"),  # Polyline
            ],
            [
                State(self.ids("surfaces-checklist"), "value"),  # surfacefiles
                State(self.ids("surfaces-de-checklist"), "value"),  # surfacefiles keys
                State(self.ids("well-settings-checklist"), "value"),  # Well settings
            ],
        )
        def _render_xsection(
            n_apply_sfc,
            n_apply_well,
            wellfile,
            polyline,
            surfacefiles,
            de_keys,
            well_settings,
        ):
            """Renders cross section view from wellfile or polyline drawn in map view"""
            _ = n_apply_sfc, n_apply_well
            ctx = callback_context
            if wellfile in self.wellfiles:
                well = self.wells[wellfile]
                is_planned = False
            else:
                well = self.planned_wells[wellfile]
                is_planned = True
            well.create_relative_hlen()
            if (
                ctx.triggered[0]["prop_id"]
                == self.ids("layered-map") + ".polyline_points"
                and polyline is not None
            ):
                well = None
            self.xsec.set_de_and_surface_lines(surfacefiles, de_keys, well, polyline)
            self.xsec.set_xsec_fig(
                surfacefiles, de_keys, well_settings, well, is_planned=is_planned
            )
            return self.xsec.fig

        @app.callback(
            Output(self.ids("well-dropdown"), "value"),  # Wellfile
            [Input(self.ids("layered-map"), "clicked_shape")],  # Clicked circle on map
        )
        def _update_render_map_and_xsection(clicked_shape):
            """Updates well dropdown value when choosing a new circle"""
            return (
                clicked_shape["id"] if clicked_shape is not None else self.wellfiles[0]
            )

        @app.callback(
            Output(self.ids("surfaces-checklist"), "value"),
            [Input(self.ids("all-surfaces-checkbox"), "value")],
        )
        def _update_surface_tickboxes(all_surfaces_checkbox):
            """Toggle on/off all surfaces in graph settings modal"""
            return self.surfacenames if all_surfaces_checkbox == ["True"] else []

        @app.callback(
            Output(self.ids("modal-graph-settings"), "is_open"),
            [
                Input(self.ids("button-open-graph-settings"), "n_clicks"),
                Input(self.ids("button-close-graph-settings"), "n_clicks"),
                Input(self.ids("button-open-graph-settings"), "disabled"),
            ],
            [State(self.ids("modal-graph-settings"), "is_open")],
        )
        def _toggle_modal_graph_settings(n_open, n_close, disabled, is_open):
            """Open or close graph settings modal button"""
            if disabled:
                switch = False
            elif n_open or n_close:
                switch = not is_open
            else:
                switch = is_open
            return switch

        @app.callback(
            Output(self.ids("surfaces-de-checklist"), "options"),
            [Input(self.ids("surfaces-checklist"), "value")],
            [State(self.ids("surfaces-de-checklist"), "options")],
        )
        def _disable_error_checkboxes(surface_values, de_options):
            """Removes ability to toggle depth error when
            corresponding surface is disabled in graph settings modal
            """
            for i, opt in enumerate(de_options):
                if (surface_values is None) or (opt["value"] not in surface_values):
                    de_options[i]["disabled"] = True
                else:
                    de_options[i]["disabled"] = False
            return de_options

        @app.callback(
            Output(self.ids("well-settings-checklist"), "value"),
            [Input(self.ids("all-well-settings-checkbox"), "value")],
        )
        def _update_well_settings_tickboxes(all_well_attributes_checkbox):
            """Toggle on/off all options in well settings modal"""
            return (
                ["zonelog", "zonation_points", "conditional_points"]
                if all_well_attributes_checkbox == ["True"]
                else []
            )

        @app.callback(
            Output(self.ids("modal-well-settings"), "is_open"),
            [
                Input(self.ids("button-open-well-settings"), "n_clicks"),
                Input(self.ids("button-close-well-settings"), "n_clicks"),
            ],
            [State(self.ids("modal-well-settings"), "is_open")],
        )
        def _toggle_modal_well_settings(n_open, n_close, is_open):
            """Open or close well settings modal button"""
            if n_open or n_close:
                return not is_open
            return is_open

        @app.callback(
            Output(self.ids("well-points-table-container"), "children"),
            [
                Input(self.ids("button-apply-columnlist"), "n_clicks"),
            ],
            [
                State(self.ids("columns-checklist"), "value"),
            ],  # columns list
        )
        # pylint: disable=unused-argument
        def display_output(n_clicks, column_list):
            """Renders wellpoints table from csv file"""
            wellpoints_df = self.df_well_target_points.update_wellpoints_df(column_list)
            return html.Div(
                [
                    dash_table.DataTable(
                        id=self.ids("well-points-table"),
                        columns=[{"name": i, "id": i} for i in wellpoints_df.columns],
                        data=wellpoints_df.to_dict("records"),
                        sort_action="native",
                        filter_action="native",
                    ),
                ]
            )

        @app.callback(
            Output(self.ids("modal-table-settings"), "is_open"),
            [
                Input(self.ids("button-open-table-settings"), "n_clicks"),
                Input(self.ids("button-close-table-settings"), "n_clicks"),
            ],
            [State(self.ids("modal-table-settings"), "is_open")],
        )
        def _toggle_modal_table_settings(n_open, n_close, is_open):
            """Open or close table settings modal button"""
            if n_open or n_close:
                return not is_open
            return is_open

        @app.callback(
            [
                Output(self.ids("hidden-div-map-view"), "hidden"),
                Output(self.ids("hidden-div-table-view"), "hidden"),
            ],
            [Input(self.ids("map-table-radioitems"), "value")],
        )
        def _toggle_left_flexbox_content(value):
            """Returns which left flexbox content is visible/hidden"""
            if value == "table-view":
                switch = True, False
            else:
                switch = False, True
            return switch

        @app.callback(
            Output(self.ids("uncertainty-table"), "data"),
            [Input(self.ids("well-dropdown"), "value")],
        )
        def _render_uncertainty_table(wellfile):
            if wellfile in self.wellfiles:
                well = self.wells[wellfile]
            else:
                well = self.planned_wells[wellfile]
            df = self.xsec.get_intersection_dataframe(well)
            return df.to_dict("records")

        @app.callback(
            Output(self.ids("surface-picks-label"), "children"),
            [Input(self.ids("well-dropdown"), "value")],
        )
        def _render_surface_picks_label(wellfile):
            if wellfile in self.wellfiles:
                wellname = self.wells[wellfile].wellname
            else:
                wellname = self.planned_wells[wellfile].wellname
            return f"Surface picks for {wellname}"

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        functions = [(get_path, [{"path": Path(fn)}]) for fn in self.allfiles]
        functions.append((find_files, [{"folder": self.basedir}]))
        if self.planned_wells_dir is not None:
            functions.append(
                (
                    find_files,
                    [
                        {"folder": self.planned_wells_dir, "pattern": "*.txt"},
                        {
                            "folder": self.basedir / "input" / "welldata",
                            "pattern": "*.txt",
                        },
                    ],
                )
            )
        functions.append(
            (
                get_surfaces,
                [
                    {
                        "basedir": self.basedir,
                        "modelfile": self.modelfile_path,
                    }
                ],
            )
        )
        return functions


@webvizstore
def get_path(path) -> Path:
    return Path(path)


def get_color(i):
    """Create a list of colors for surface layers
    Args:
        i: Index of surface layer in surfacefiles list
    Returns:
        List of colors for surface layers
    """
    colors = [
        "rgb(70,130,180)",  # Steel blue
        "rgb(0,0,255)",  # Blue
        "rgb(173,255,47)",  # Green yellow
        "rgb(0,128,0)",  # Green
        "rgb(0,255,0)",  # Lime
        "rgb(60,179,113)",  # Medium sea green
        "rgb(255,105,180)",  # Pink
        "rgb(221,160,221)",  # Plum
        "rgb(255,255,0)",  # Yellow
        "rgb(244,164,96)",  # Tan
        "rgb(255,140,0)",  # Orange
        "rgb(255,69,0)",  # Blood orange
        "rgb(255,0,0)",  # Red
        "rgb(220,20,60)",  # Crimson
        "rgb(255,0,255)",  # Fuchsia
    ]
    n_colors = len(colors)
    return colors[(i) % (n_colors)]


def load_surfaces(basedir: Path, modelfile):
    surface_types = {
        "depth": "d_",
        "depth-trend": "dt_",
        "depth-error": "de_",
        "depth-trend-error": "dte_",
        "depth-residual": "dr_",
    }
    surface_data = json.load(get_surfaces(basedir, modelfile))
    for surface in surface_data:
        for stype in surface_types.values():
            surface[stype] = surface_from_json(surface[stype]) if stype else None
    return surface_data


@webvizstore
def get_surfaces(basedir: Path, modelfile) -> io.BytesIO:
    modelfile = ET.parse(get_path(modelfile)).getroot()
    surface_types = {
        "depth": "d_",
        "depth-trend": "dt_",
        "depth-error": "de_",
        "depth-trend-error": "dte_",
        "depth-residual": "dr_",
    }
    surface_wrappers = modelfile.findall(".//surface")
    surfaces = []
    for element in surface_wrappers:
        surface = {
            "name": element.findtext("name"),
            "topofzone": element.findtext("top-of-zone"),
        }
        for s_mtype, s_type in surface_types.items():
            surface[s_type] = (
                surface_to_json(
                    str(
                        basedir
                        / "output"
                        / "surfaces"
                        / f"{s_type}{surface['name']}.rxb"
                    )
                )
                if element.find("output").findtext(s_mtype) == "yes"
                else None
            )
        surfaces.append(surface)
    return io.BytesIO(json.dumps(surfaces).encode())


def surface_to_json(surfacepath: Path) -> str:
    surface = xtgeo.surface_from_file(str(surfacepath), fformat="irap_binary")
    return json.dumps(
        {
            "ncol": surface.ncol,
            "nrow": surface.nrow,
            "xori": surface.xori,
            "yori": surface.yori,
            "rotation": surface.rotation,
            "xinc": surface.xinc,
            "yinc": surface.yinc,
            "values": surface.values.copy().filled(np.nan).tolist(),
        }
    )


def surface_from_json(surfaceobj):
    data = json.loads(surfaceobj)
    surface = xtgeo.RegularSurface(**data)
    surface.values = np.array(data["values"])
    return surface


# TODO(Sigurd) Delete unused function?
# def extract_topofzone_names(modelfile):
#    modelfile = ET.parse(modelfile).getroot()
#    surface_wrappers = modelfile.findall(".//surface")
#    topofzone_names = []
#    for element in surface_wrappers:
#        name = element.findtext("top-of-zone")
#        topofzone_names.append(name)
#    return topofzone_names


# TODO(Sigurd) Delete unused function?
# def get_well_files(basedir: Path) -> io.BytesIO:
#    well_dir = os.path.join(basedir, "input", "welldata")
#    well_dir = basedir / "input" / "welldata"
#    well_files = []
#    try:
#        for file in os.listdir(well_dir):
#            if Path(file).suffix == ".txt":
#                well_files.append(os.path.join(well_dir, file))
#    except FileNotFoundError:
#        pass
#    well_files.sort()
#    return io.BytesIO(json.dumps(well_files).encode())


def get_target_points(basedir: Path) -> Path:
    return get_path(basedir / "output" / "log_files" / "targetpoints.csv")


def get_well_points(basedir: Path) -> Path:
    return get_path(basedir / "output" / "log_files" / "wellpoints.csv")


def get_zonelog_name(modelfile):
    modelfile = ET.parse(modelfile).getroot()
    zonelog_wrapper = modelfile.findtext(".//zone-log-name")
    return zonelog_wrapper


def get_zonation_status(basedir: Path) -> Path:
    return get_path(basedir / "output" / "log_files" / "zonation_status.csv")


# TODO(Sigurd) Delete unused?
# @webvizstore
# def get_surface_files(basedir: Path, surface_names, surface_type) -> io.BytesIO:
#    surface_dir = basedir / "output" / "surfaces"
#    surface_files = [
#        surface_dir / f"{surface_type}{surface_name}.rxb"
#        for surface_name in surface_names
#    ]
#    for path in surface_files:
#        if not path.is_file():
#            surface_files = None
#    if surface_files is not None:
#        surface_files = [str(surf) for surf in surface_files]
#    return io.BytesIO(json.dumps(surface_files).encode())


@webvizstore
def find_files(folder: Path, pattern: str = "*") -> io.BytesIO:
    return io.BytesIO(
        json.dumps(
            sorted(
                [
                    str(filename)
                    for filename in folder.glob(f"**/{pattern}")
                    if filename.is_file()
                ]
            )
        ).encode()
    )



from typing import Any, Callable, Dict, List, Tuple

from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import StrEnum

from webviz_subsurface._models import StratigraphyModel, WellAttributesModel

from ..._providers import EnsembleTableProviderFactory
from ._utils import WellCompletionDataModel
from ._views._well_completion_view import WellCompletionView, WellCompletionViewElement


class WellCompletion(WebvizPluginABC):
    """Vizualizes Eclipse well completions data per well. The data is grouped per well
    and zone and can be filtered according to flexible well categories.

    ---

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`wellcompletiondata_file`:** `.arrow` file with well completion data
    * **`stratigraphy_file`:** `.json` file defining the stratigraphic levels
    * **`well_attributes_file`:** `.json` file with categorical well attributes
    * **`kh_unit`:** e.g. mD·m, will try to extract from eclipse files if defaulted
    * **`kh_decimal_places`:**

    ---
    The minimum requirement is to define `ensembles`.

    **Well Completion data**

    `wellcompletiondata_file` is a path to an `.arrow` file stored per realization (e.g in \
    `share/results/tables/wellcompletiondata.arrow`). This file can be exported to disk by using the
    `ECL2CSV` forward model in ERT with subcommand `wellcompletiondata`. This forward model will
    read the eclipse `COMPDAT`, but then aggregate from layer to zone according to a given zone ➔
    layer mapping `.lyr` file. If the `use_wellconnstatus` option is used, then the `OP/SH` status
    of each well connection is deduced from `CPI` summary data which in som cases is more accurate
    than the data coming from parsing the schedule file (f.ex if connections are changing status in
    an `ACTION` keyword).

    The reason for doing the layer ➔ zone aggregation in the FMU post processing instead of in the
    plugin is to improve speed and to open up for more functionality.

    **Stratigraphy file**

    The `stratigraphy_file` file is intended to be generated per realization by an internal \
    RMS script as part of the FMU workflow, but can also be set up manually and copied to each
    realization. The stratigraphy is a tree structure, where each node has a name, an optional
    `color` parameter, and an optional `subzones` parameter which itself is a list of the same
    format.
    ```json
    [
        {
            "name": "ZoneA",
            "color": "#FFFFFF",
            "subzones": [
                {
                    "name": "ZoneA.1"
                },
                {
                    "name": "ZoneA.2"
                }
            ]
        },
        {
            "name": "ZoneB",
            "color": "#FFF000",
            "subzones": [
                {
                    "name": "ZoneB.1",
                    "color": "#FFF111"
                },
                {
                    "name": "ZoneB.2",
                    "subzones: {"name": "ZoneB.2.2"}
                }
            ]
        },
    ]
    ```
    The `stratigraphy_file` and the `zone_layer_mapping_file` will be combined to create the final \
    stratigraphy. A node will be removed if the name or any of the subnode names are not \
    present in the zone layer mapping. A Value Error is raised if any zones are present in the
    zone layer mapping but not in the stratigraphy.

    Colors can be supplied both trough the stratigraphy and through the zone_layer_mapping. \
    The following prioritization will be applied:
    1. Colors specified in the stratigraphy
    2. Colors specified in the zone layer mapping lyr file
    3. If none of the above is specified, theme colors will be added to the leaves of the tree

    **Well Attributes file**

    The `well_attributes_file` file is intended to be generated per realization by an internal \
    RMS script as part of the FMU workflow. A sample script will be made available, but it is \
    possible to manually set up the file and copy it to the correct folder on the scratch disk.\
    The categorical well attributes are completely flexible.

    The file should be a `.json` file on the following format:
    ```json
    {
        "version" : "0.1",
        "wells" : [
            {
                "alias" : {
                    "eclipse" : "OP_1"
                },
                "attributes" : {
                    "mlt_singlebranch" : "mlt",
                    "structure" : "East",
                    "welltype" : "producer"
                },
                "name" : "OP_1"
            },
            {
                "alias" : {
                    "eclipse" : "GI_1"
                },
                "attributes" : {
                    "mlt_singlebranch" : "singlebranch",
                    "structure" : "West",
                    "welltype" : "gas injector"
                },
                "name" : "GI_1"
            },
        ]
    }
    ```

    **KH unit**

    If defaulted, the plugin will look for the unit system of the Eclipse deck in the DATA file. \
    The kh unit will be deduced from the unit system, e.g. mD·m if METRIC.

    """

    class Ids(StrEnum):
        WELL_COMPLETION_VIEW = "well-completion-view"

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: list,
        wellcompletiondata_file: str = "share/results/tables/wellcompletiondata.arrow",
        stratigraphy_file: str = "rms/output/zone/stratigraphy.json",
        well_attributes_file: str = "rms/output/wells/well_attributes.json",
    ):

        super().__init__(stretch=True)
        factory = EnsembleTableProviderFactory.instance()

        self._data_models = {}
        for ens_name in ensembles:
            ens_path = webviz_settings.shared_settings["scratch_ensembles"][ens_name]
            self._data_models[ens_name] = WellCompletionDataModel(
                ensemble_path=ens_path,
                wellcompletion_provider=factory.create_from_per_realization_arrow_file(
                    ens_path, wellcompletiondata_file
                ),
                stratigraphy_model=StratigraphyModel(
                    ens_name, ens_path, stratigraphy_file
                ),
                well_attributes_model=WellAttributesModel(
                    ens_name, ens_path, well_attributes_file
                ),
                theme_colors=webviz_settings.theme.plotly_theme["layout"]["colorway"],
            )

        self.add_view(
            WellCompletionView(self._data_models),
            self.Ids.WELL_COMPLETION_VIEW,
        )

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        return [
            webviz_store_tuple
            for _, ens_data_model in self._data_models.items()
            for webviz_store_tuple in ens_data_model.webviz_store
        ]

    @property
    def tour_steps(self) -> List[Dict[str, Any]]:
        return [
            {
                "id": self.view(self.Ids.WELL_COMPLETION_VIEW)
                .settings_group(WellCompletionView.Ids.SETTINGS)
                .get_unique_id(),
                "content": "Menu for selecting ensemble and other options",
            },
            {
                "id": self.view(self.Ids.WELL_COMPLETION_VIEW)
                .view_element(WellCompletionView.Ids.VIEW_ELEMENT)
                .component_unique_id(WellCompletionViewElement.Ids.COMPONENT),
                "content": "Visualization of the well completions. "
                "Time slider for selecting which time steps to display. "
                "Different vizualisation and filtering alternatives are available "
                "in the upper right corner.",
            },
        ]



import abc
from typing import List, Optional, Sequence

import pandas as pd


class DerivedVectorsAccessor:
    def __init__(self, accessor_realizations: List[int]) -> None:
        self._accessor_realizations: List[int] = accessor_realizations

    @abc.abstractmethod
    def has_provider_vectors(self) -> bool:
        ...

    @abc.abstractmethod
    def has_per_interval_and_per_day_vectors(self) -> bool:
        ...

    @abc.abstractmethod
    def has_vector_calculator_expressions(self) -> bool:
        ...

    @abc.abstractmethod
    def get_provider_vectors_df(
        self, realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        ...

    @abc.abstractmethod
    def create_per_interval_and_per_day_vectors_df(
        self,
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        ...

    @abc.abstractmethod
    def create_calculated_vectors_df(
        self, realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        ...

    def create_valid_realizations_query(
        self, selected_realizations: List[int]
    ) -> Optional[List[int]]:
        """Create realizations query for accessor based on selected realizations.

        `Returns:`
        - None - If all realizations for accessor is selected, i.e. the query is non-filtering
        - List[int] - List of realization numbers existing for the accessor - empty list
        is returned if no realizations exist.
        """
        if set(self._accessor_realizations).issubset(set(selected_realizations)):
            return None
        return [
            realization
            for realization in selected_realizations
            if realization in self._accessor_realizations
        ]



class PluginIds:
    # pylint: disable=too-few-public-methods
    class Stores:
        # pylint: disable=too-few-public-methods
        SELECTED_ENSEMBLES = "selected-ensembles"
        SELECTED_DATES = "selected-dates"
        SELECTED_PHASE = "selected-phase"
        SELECTED_WELLS = "selected-wells"
        SELECTED_COMBINE_WELLS_COLLECTION = "selected-combine-wells-collection"
        SELECTED_WELL_COLLECTIONS = "selected-well-collections"
        SELECTED_REALIZATIONS = "selected-realizations"

        SELECTED_FIG_LAYOUT_HEIGHT = "selected-fig-layout-height"

    class SharedSettings:
        # pylint: disable=too-few-public-methods
        FILTER = "filter"

    class MisfitViews:
        # pylint: disable=too-few-public-methods
        GROUP_NAME = "Misfit analysis"

        PRODUCTION_MISFIT_PER_REAL = "production-misfit-per-real"
        WELL_PRODUCTION_COVERAGE = "well-production-coverage"
        WELL_PRODUCTION_HEATMAP = "well-production-heatmap"



from typing import List, Union

import numpy as np
import pandas as pd


def calc_from_cumulatives(
    data: pd.DataFrame,
    column_keys: Union[List[str], str],
    time_index: str,
    time_index_input: str,
    as_rate: bool,
) -> pd.DataFrame:
    """Calculates interval delta and average rate at given time interval `time_index`.
    `data` is a dataframe which has the columns "DATE", "ENSEMBLE", "REAL" and the columns
    that you want to make the calculation for

    Assumes that the data is already sampled to a time interval `time_index_input`.
    `time_index` can currently be `daily`, `monthly` or `yearly`.

    Interpolation is not performed, and the time interval `time_index` therefore has to be shorter
    or equal to `time_index_input` (e.g. `time_index` = `yearly` is compatible with
    `time_index_input` = `monthly`, but the opposite is invalid).

    `column_keys` define vectors to take calculate for of.

    If `as_rate` == True, the calculated cumulative for an interval is divided by the number of
    days in the interval to give a rate per day.

    Assumes that the data for each realization comes in consecutive lines, and that it is sorted as
    increasing in time.

    The average rate and interval production is stored at the beginning date of the interval,
    opposite to rates in e.g. the Eclipse simulator's summary format, but similar to fmu-ensemble's
    get_volumetric_rates(). E.g. for a monthly interval, the values stored at 2010-01-01 would be
    the average rate and production for January 2010.
    """
    data = data.copy()
    if isinstance(column_keys, str):
        column_keys = [column_keys]

    # Converting the DATE axis to datetime to allow for timedeltas
    data.loc[:, ["DATE"]] = pd.to_datetime(data["DATE"])
    _verify_time_index(data, time_index, time_index_input)
    # Creating a column of unique values per ensemble-realization combination. A non-zero
    # diff of this column will then mean that it is a diff between different realizations.
    # Could alternatively loop over ensembles and realizations, but this is quicker for
    # larger datasets.
    data["ensrealuid"] = (
        data["ENSEMBLE"].astype("category").cat.codes * data["REAL"].nunique()
        + data["REAL"]
    )

    # Allows us to resample on DATE without merging ensembles and realizations.
    data.set_index(["ENSEMBLE", "REAL", "DATE"], inplace=True)

    # Resample on DATE frequency
    data = _resample_time_index(
        df=data, time_index=time_index, time_index_input=time_index_input
    )

    data.reset_index(level=["ENSEMBLE", "REAL"], inplace=True)

    calc_cols = {vec: rename_vec_from_cum(vec, as_rate) for vec in column_keys}
    listed_calc_cols = [calc_cols[col] for col in column_keys]

    # Take diff of given column_keys + the ensemble-realization identifier.
    # Preserve ENSEMBLE and REAL
    diff_cum = pd.concat(
        [
            data[["ENSEMBLE", "REAL"]],
            data[["ensrealuid"] + column_keys]
            .diff()
            .shift(-1)
            .rename(
                mapper=calc_cols,
                axis=1,
            ),
        ],
        axis=1,
    )
    diff_cum[listed_calc_cols] = diff_cum[listed_calc_cols].fillna(value=0)

    # Reset index (DATE becomes regular column)
    diff_cum.reset_index(inplace=True)

    # Convert interval cumulative to daily average rate if requested
    if as_rate:
        days = diff_cum["DATE"].diff().shift(-1).dt.days.fillna(value=0)
        for vec in column_keys:
            with np.errstate(invalid="ignore"):
                diff_cum.loc[:, calc_cols[vec]] = (
                    diff_cum[calc_cols[vec]].values / days.values
                )

    # Set last value of each real to 0 (as we don't loop over the realizations)
    diff_cum.loc[diff_cum["ensrealuid"] != 0, listed_calc_cols] = 0
    diff_cum.drop("ensrealuid", axis=1, inplace=True)

    return diff_cum


def _verify_time_index(
    df: pd.DataFrame, time_index: str, time_index_input: str
) -> None:
    freqs = {"D": "daily", "MS": "monthly", "AS-JAN": "yearly"}
    valid_time_indices = {
        "daily": ["daily", "monthly", "yearly"],
        "monthly": ["monthly", "yearly"],
        "yearly": ["yearly"],
    }
    inferred_frequency = pd.infer_freq(sorted(df["DATE"].unique()))
    if not freqs.get(inferred_frequency) == time_index_input:
        raise ValueError(
            "The DataFrame most likely contains data points which are not sampled on "
            f"frequency time_index_input={time_index_input}. The inferred frequency from the "
            f"unique DATE values was {inferred_frequency}."
        )
    if not time_index in valid_time_indices[time_index_input]:
        raise ValueError(
            f"The time_index {time_index} has higher frequency than time_index_input "
            f"{time_index_input}. Valid time_index options are "
            f"{valid_time_indices[time_index_input]}."
        )


def _resample_time_index(
    df: pd.DataFrame, time_index: str, time_index_input: str
) -> pd.DataFrame:
    if time_index == time_index_input:
        return df
    if time_index == "yearly" and time_index_input in ["daily", "monthly"]:
        return df.groupby(
            [
                pd.Grouper(level="ENSEMBLE"),
                pd.Grouper(level="REAL"),
                pd.Grouper(level="DATE", freq="YS"),
            ]
        ).first()
    if time_index == "monthly" and time_index_input == "daily":
        return df.groupby(
            [
                pd.Grouper(level="ENSEMBLE"),
                pd.Grouper(level="REAL"),
                pd.Grouper(level="DATE", freq="MS"),
            ]
        ).first()
    raise ValueError(
        f"Cannot combine `time_index`={time_index} and `time_index_input`={time_index_input}. "
        "Ensure that `time_index_input` has a higher than or equal frequency to `time_index`. "
        "Valid options are `daily`, `monthly` and `yearly` (dependent on input data)."
    )


def rename_vec_from_cum(vector: str, as_rate: bool) -> str:
    """This function assumes that it is a cumulative/total vector named in the Eclipse standard
    and is fairly naive when converting to rate. Based in the list in libecl
    https://github.com/equinor/libecl/blob/69f1ee0ddf696c87b6d85eca37eed7e8b66ac2db/\
        lib/ecl/smspec_node.cpp#L531-L586
    the T identifying total/cumulative should not occur before letter 4,
    as all the listed strings are prefixed with one or two letters in the vectors.
    Therefore starting the replace at the position 3 (4th letter) to reduce risk of errors
    in the conversion to rate naming, but hard to be completely safe.
    """
    return (
        f"AVG_{vector[0:3] + vector[3:].replace('T', 'R', 1)}"
        if as_rate
        else f"INTVL_{vector}"
    )



from typing import Callable, List, Union

import webviz_core_components as wcc
import webviz_subsurface_components as wsc
from dash import html

from ..models import (
    PropertyStatisticsModel,
    ProviderTimeSeriesDataModel,
    SimulationTimeSeriesModel,
)


def vector_selector(
    get_uuid: Callable,
    vector_model: Union[SimulationTimeSeriesModel, ProviderTimeSeriesDataModel],
) -> wsc.VectorSelector:
    return wsc.VectorSelector(
        id=get_uuid("property-response-vector-select"),
        maxNumSelectedNodes=1,
        data=vector_model.vector_selector_data,
        persistence=True,
        persistence_type="session",
        selectedTags=vector_model.vectors[:1],
        numSecondsUntilSuggestionsAreShown=0.5,
        lineBreakAfterTag=True,
    )


def ensemble_selector(
    get_uuid: Callable,
    ensembles: List[str],
    tab: str,
    multi: bool = False,
    initial_ensemble: str = None,
) -> html.Div:
    return wcc.Dropdown(
        label="Ensemble",
        id={"id": get_uuid("ensemble-selector"), "tab": tab},
        options=[{"label": ens, "value": ens} for ens in ensembles],
        multi=multi,
        value=initial_ensemble if initial_ensemble is not None else ensembles[0],
        clearable=False,
    )


def delta_ensemble_selector(
    get_uuid: Callable, ensembles: List[str], tab: str, multi: bool = False
) -> html.Div:
    return wcc.Dropdown(
        label="Delta Ensemble",
        id={"id": get_uuid("delta-ensemble-selector"), "tab": tab},
        options=[{"label": ens, "value": ens} for ens in ensembles],
        multi=multi,
        value=ensembles[-1],
        clearable=False,
    )


def property_selector(
    get_uuid: Callable,
    properties: List[str],
    tab: str,
    multi: bool = False,
    initial_property: str = None,
) -> html.Div:
    display = "none" if len(properties) < 2 else "inline"
    return html.Div(
        style={"display": display},
        children=[
            wcc.Dropdown(
                label="Property",
                id={"id": get_uuid("property-selector"), "tab": tab},
                options=[{"label": prop, "value": prop} for prop in properties],
                multi=multi,
                value=initial_property
                if initial_property is not None
                else properties[0],
                clearable=False,
            )
        ],
    )


def source_selector(
    get_uuid: Callable,
    sources: List[str],
    tab: str,
    multi: bool = False,
    initial_source: str = None,
) -> html.Div:
    display = "none" if len(sources) < 2 else "inline"
    return html.Div(
        style={"display": display},
        children=[
            wcc.Dropdown(
                label="Source",
                id={"id": get_uuid("source-selector"), "tab": tab},
                options=[{"label": source, "value": source} for source in sources],
                multi=multi,
                value=initial_source if initial_source is not None else sources[0],
                clearable=False,
            ),
        ],
    )


def make_filter(
    get_uuid: Callable,
    tab: str,
    df_column: str,
    column_values: list,
    multi: bool = True,
    value: Union[str, float] = None,
) -> html.Div:
    return wcc.SelectWithLabel(
        label=df_column.lower().capitalize(),
        id={
            "id": get_uuid("filter-selector"),
            "tab": tab,
            "selector": df_column,
        },
        options=[{"label": i, "value": i} for i in column_values],
        value=[value] if value is not None else column_values,
        multi=multi,
        size=min(20, len(column_values)),
    )


def filter_selector(
    get_uuid: Callable,
    property_model: PropertyStatisticsModel,
    tab: str,
    multi: bool = True,
    value: Union[str, float] = None,
) -> html.Div:
    return html.Div(
        children=[
            html.Div(
                children=[
                    make_filter(
                        get_uuid=get_uuid,
                        tab=tab,
                        df_column=sel,
                        column_values=property_model.selector_values(sel),
                        multi=multi,
                        value=value,
                    )
                    for sel in property_model.selectors
                ]
            ),
        ]
    )



from dash import dcc, html
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewElementABC
from webviz_subsurface_components import DeckGLMap


class VTKView3D(ViewElementABC):
    class Ids(StrEnum):
        ID = "vtk"
        VIEW = "vtk-view"

        INFOBOX = "info-box"
        ACTUALVALUERANGE = "actual-value-range"

    def __init__(self) -> None:
        super().__init__()

    def inner_layout(self) -> Component:

        return html.Div(
            style={"position": "relative", "height": "90vh"},
            children=[
                html.Div(
                    style={"zIndex": 10000, "position": "absolute", "top": 0},
                    children=[
                        html.Div(
                            children=[
                                html.Div(
                                    id=self.register_component_unique_id(
                                        VTKView3D.Ids.INFOBOX
                                    ),
                                    style={
                                        "border": "black",
                                        "margin": "10px",
                                        "padding": "5px",
                                        "fontSize": "0.8em",
                                        "borderRadius": "10px",
                                        "backgroundColor": "lightgrey",
                                    },
                                ),
                            ]
                        ),
                    ],
                ),
                html.Div(
                    style={"position": "absolute", "width": "100%", "height": "90%"},
                    children=[
                        DeckGLMap(
                            id=self.register_component_unique_id(VTKView3D.Ids.VIEW),
                            layers=[
                                {
                                    "@@type": "AxesLayer",
                                    "id": "axes-layer",
                                    "bounds": [
                                        0,
                                        0,
                                        -1,
                                        1,
                                        1,
                                        0,
                                    ],
                                },
                                {
                                    "@@type": "Grid3DLayer",
                                    "id": "Grid3DLayer",
                                    "material": True,
                                    "colorMapName": "Physics reverse",
                                    "scaleZ": 1,
                                    "pointsUrl": "/grid/points/test",
                                    "polysUrl": "/grid/polys/test",
                                    "propertiesUrl": "/grid/scalar/test",
                                },
                            ],
                            views={
                                "layout": [1, 1],
                                "viewports": [{"id": "view_1", "show3D": True}],
                            },
                        ),
                    ],
                ),
                html.Div(
                    style={"position": "absolute", "bottom": 0},
                    children=html.Label(
                        children="⚠️ This is an experimental plugin that might be removed "
                        "on short notice. Verify all results in ResInsight or RMS.",
                    ),
                ),
                dcc.Store(
                    id=self.register_component_unique_id(
                        VTKView3D.Ids.ACTUALVALUERANGE
                    ),
                    data=[0, 0],
                ),
            ],
        )



import abc
from typing import Dict, List, Optional

import pandas as pd
import plotly.graph_objects as go

from .._types import FanchartOptions, StatisticsOptions


class GraphFigureBuilderBase(abc.ABC):
    """
    Base class for creating/building serializable Output property data
    for the callback. Has functionality for creating various plot traces, where
    the class inheriting the base is responsible to retrieve the data and place
    correct in graph figure - e.g. place traces in correct subplots, set correct
    titles, legends and so on.

    Contains interface for adding graph data and retrieving the serialized data
    for callback Output property.

    Contains self._figure, an empty FigureWidget to either use or override
    """

    def __init__(self) -> None:
        self._figure = go.Figure()

    # ------------------------------------
    #
    # Public functions
    #
    # ------------------------------------

    def get_serialized_figure(self) -> dict:
        """
        Get the built figure on a JSON serialized format - i.e. a dictionary
        """
        return self._figure.to_dict()

    @abc.abstractmethod
    def create_graph_legends(self) -> None:
        """Create legends for graphs after trace data is added"""

    @abc.abstractmethod
    def add_realizations_traces(
        self,
        vectors_df: pd.DataFrame,
        ensemble: str,
        color_lightness_scale: Optional[float] = None,
    ) -> None:
        """Add realization traces to figure

        `Input:`
        * vectors_df: pd.Dataframe - Dataframe with columns:
            ["DATE", "REAL", vector1, ..., vectorN]

        * ensemble: str - Name of ensemble providing the input vector data
        * color_lightness_scale: float - Color lightness scale percentage, to adjust trace colors
        relative to original trace line color defined. Range 50% - 150%
        """

    @abc.abstractmethod
    def add_statistics_traces(
        self,
        vectors_statistics_df: pd.DataFrame,
        ensemble: str,
        statistics_options: List[StatisticsOptions],
        line_width: Optional[int] = None,
        color_lightness_scale: Optional[float] = None,
    ) -> None:
        """Add statistics traces to figure

        `Input:`
        * vector_statistics_df: pd.Dataframe - Dataframe with double column level:\n
          [ "DATE",    vector1,                        ... vectorN
                       MEAN, MIN, MAX, P10, P90, P50   ... MEAN, MIN, MAX, P10, P90, P50]

        * ensemble: str - Name of ensemble providing the input vector data
        * statistics_options: List[StatisticsOptions] - List of statistics options traces to include
        * line_width: int - Line width for statistics traces
        * color_lightness_scale: float - Color lightness scale percentage, to adjust trace colors
        relative to original trace line color defined. Range 50% - 150%
        """

    @abc.abstractmethod
    def add_fanchart_traces(
        self,
        vectors_statistics_df: pd.DataFrame,
        ensemble: str,
        fanchart_options: List[FanchartOptions],
    ) -> None:
        """
        Add fanchart traces for vectors in provided vectors statistics dataframe

        `Input:`
        * vector_statistics_df: pd.Dataframe - Dataframe with double column level:\n
          [ "DATE",    vector1,                        ... vectorN
                       MEAN, MIN, MAX, P10, P90, P50   ... MEAN, MIN, MAX, P10, P90, P50]

        * ensemble: str - Name of ensemble providing the input vector data
        * fanchart_options: List[StatisticsOptions] - List of fanchart options traces to include
        """

    @abc.abstractmethod
    def add_history_traces(
        self,
        vectors_df: pd.DataFrame,
        ensemble: str,
    ) -> None:
        """Add traces for historical vectors in dataframe columns

        `Input:`
        * vectors_df: pd.Dataframe - Dataframe with non-historical vector names in columns and their
        historical data in rows. With columns:\n
            ["DATE", "REAL", vector1, ..., vectorN]

        * ensemble: str - Name of ensemble providing the input vector data
        """

    @abc.abstractmethod
    def add_vector_observations(
        self, vector_name: str, vector_observations: dict
    ) -> None:
        """Add traces for vector observations

        `Input:`
        * vector_name: str - Vector to add observations for
        * vector_observations: dict - Dictionary with observation data for vector
        """

    # ------------------------------------
    #
    # Private functions
    #
    # ------------------------------------

    @abc.abstractmethod
    def _add_vector_traces_set_to_figure(
        self, vector_traces_set: Dict[str, List[dict]], ensemble: Optional[str] = None
    ) -> None:
        """
        Add list of vector line traces to figure.

        Places line traces for specified vector into correct subplot of figure

        `Input:`
        * vector_traces_set: Dict[str, List[dict]] - Dictionary with vector names and list
        of vector line traces for figure.
        * ensemble: str - Optional name of ensemble providing the input vector data
        """

    @abc.abstractmethod
    def _add_vector_trace_set_to_figure(
        self, vector_trace_set: Dict[str, dict], ensemble: Optional[str] = None
    ) -> None:
        """
        Add vector line trace to figure

        Places line trace for specified vector into correct subplot of figure

        `Input:`
        * vector_trace_set: Dict[str, dict] - Dictionary with vector name and single
        vector line trace for figure.
        * ensemble: str - Optional name of ensemble providing the input vector data
        """



from dash import dash_table
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewElementABC


class TornadoTable(ViewElementABC):
    class IDs(StrEnum):
        TABLE = "table"

    def __init__(self, height: str = "75vh") -> None:
        super().__init__()
        self.height = height

    def inner_layout(self) -> dash_table.DataTable:
        return dash_table.DataTable(
            id=self.register_component_unique_id(TornadoTable.IDs.TABLE),
            style_cell={"whiteSpace": "normal", "height": "auto"},
        )



import datetime
from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

from .._utils import datetime_utils


class ResamplingFrequencySettings(SettingsGroupABC):
    class Ids(StrEnum):
        RESAMPLING_FREQUENCY_DROPDOWN = "resampling-frequency-dropdown"
        RELATIVE_DATE_DROPDOWN = "relative-date-dropdown"

    def __init__(
        self,
        disable_resampling_dropdown: bool,
        selected_resampling_frequency: Frequency,
        ensembles_dates: List[datetime.datetime],
        input_provider_set: EnsembleSummaryProviderSet,
    ) -> None:
        super().__init__("Resampling frequency")
        self._disable_resampling_dropdown = disable_resampling_dropdown
        self._selected_resampling_frequency = selected_resampling_frequency
        self._ensembles_dates = ensembles_dates
        self._input_provider_set = input_provider_set

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                id=self.register_component_unique_id(
                    ResamplingFrequencySettings.Ids.RESAMPLING_FREQUENCY_DROPDOWN
                ),
                clearable=False,
                disabled=self._disable_resampling_dropdown,
                options=[
                    {
                        "label": frequency.value,
                        "value": frequency.value,
                    }
                    for frequency in Frequency
                ],
                value=self._selected_resampling_frequency,
                style={
                    "margin-bottom": "10px",
                },
            ),
            wcc.Label(
                "Data relative to date:",
                style={
                    "font-style": "italic",
                },
            ),
            wcc.Dropdown(
                clearable=True,
                disabled=self._disable_resampling_dropdown,
                id=self.register_component_unique_id(
                    ResamplingFrequencySettings.Ids.RELATIVE_DATE_DROPDOWN
                ),
                options=[
                    {
                        "label": datetime_utils.to_str(_date),
                        "value": datetime_utils.to_str(_date),
                    }
                    for _date in sorted(self._ensembles_dates)
                ],
            ),
            wcc.Label(
                "NB: Disabled for pre-sampled data",
                style={"font-style": "italic"}
                if self._disable_resampling_dropdown
                else {"display": "none"},
            ),
        ]



from .derived_delta_ensemble_vectors_accessor_impl import (
    DerivedDeltaEnsembleVectorsAccessorImpl,
)
from .derived_ensemble_vectors_accessor_impl import DerivedEnsembleVectorsAccessorImpl
from .derived_vectors_accessor import DerivedVectorsAccessor



import glob
import logging
import os
import re
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from typing import Dict, List, Set

import pyarrow as pa

from webviz_subsurface._utils.perf_timer import PerfTimer

LOGGER = logging.getLogger(__name__)


@dataclass
class FileEntry:
    real: int
    filename: str


def _discover_arrow_unsmry_files(globpattern: str) -> List[FileEntry]:

    realidxregexp = re.compile(r"realization-(\d+)")

    globbedpaths = glob.glob(globpattern)

    visited_folders: Set[str] = set()

    file_list: List[FileEntry] = []
    for path in globbedpaths:
        this_folder = os.path.dirname(path)
        if this_folder in visited_folders:
            raise ValueError(f"Multiple matches found in folder: {this_folder}")
        visited_folders.add(this_folder)

        real = None
        for path_comp in reversed(path.split(os.path.sep)):
            realmatch = re.match(realidxregexp, path_comp)
            if realmatch:
                real = int(realmatch.group(1))
                break

        if real is None:
            raise ValueError(f"Unable to determine realization number for file: {path}")

        file_list.append(FileEntry(real=real, filename=path))

    # Sort the file entries on realization number
    file_list = sorted(file_list, key=lambda e: e.real)
    return file_list


def _load_table_from_arrow_file(entry: FileEntry) -> pa.Table:
    LOGGER.debug(f"loading table real={entry.real}: {entry.filename}")
    source = pa.memory_map(entry.filename, "r")
    reader = pa.ipc.RecordBatchFileReader(source)
    return reader.read_all()


def load_per_realization_arrow_unsmry_files(
    ens_path: str, rel_file_pattern: str
) -> Dict[int, pa.Table]:
    """Load summary data stored in per-realization arrow files.
    Returns dictionary containing a PyArrow table for each realization, indexed by
    realization number.

    `rel_file_pattern` denotes a file pattern relative to the realization's runpath,
    typical value is: "share/results/unsmry/*.arrow"
    """

    LOGGER.debug(f"load_per_realization_arrow_unsmry_files() starting - {ens_path}")
    LOGGER.debug(f"looking for .arrow files using relative pattern: {rel_file_pattern}")
    timer = PerfTimer()

    per_real_tables: Dict[int, pa.Table] = {}
    globpattern = os.path.join(ens_path, rel_file_pattern)
    files_to_process = _discover_arrow_unsmry_files(globpattern)
    if len(files_to_process) == 0:
        LOGGER.warning(f"No arrow files were discovered in: {ens_path}")
        LOGGER.warning(f"Glob pattern used: {globpattern}")
        return per_real_tables

    with ProcessPoolExecutor() as executor:
        futures = executor.map(_load_table_from_arrow_file, files_to_process)
    for i, table in enumerate(futures):
        real = files_to_process[i].real
        per_real_tables[real] = table

    # for entry in files_to_process:
    #     table = _load_table_from_arrow_file(entry)
    #     per_real_tables[entry.real] = table

    LOGGER.debug(
        f"load_per_realization_arrow_unsmry_files() "
        f"finished in: {timer.elapsed_s():.2f}s"
    )

    return per_real_tables



from typing import Callable

import webviz_core_components as wcc
from dash import dcc, html


# pylint: disable = too-few-public-methods
class LayoutElements:
    LAYOUT = "layout"
    ENSEMBLE_DROPDOWN = "ensemble_dropdown"
    WELL_COMPLETIONS_COMPONENT = "well_completions_component"


def main_layout(get_uuid: Callable, ensembles: list) -> html.Div:
    return html.Div(
        id=get_uuid(LayoutElements.LAYOUT),
        children=[
            dcc.Store(
                id=get_uuid(LayoutElements.ENSEMBLE_DROPDOWN), storage_type="session"
            ),
            wcc.FlexBox(
                children=[
                    wcc.Selectors(
                        label="Ensemble",
                        children=[
                            wcc.Dropdown(
                                id=get_uuid(LayoutElements.ENSEMBLE_DROPDOWN),
                                options=[
                                    {"label": ens, "value": ens} for ens in ensembles
                                ],
                                clearable=False,
                                value=ensembles[0],
                                persistence=True,
                                persistence_type="session",
                            ),
                        ],
                    ),
                    html.Div(style={"flex": 4}),
                ],
            ),
            html.Div(
                id=get_uuid(LayoutElements.WELL_COMPLETIONS_COMPONENT),
            ),
        ],
    )


def layout_tour_steps(get_uuid: Callable) -> list:
    return [
        {
            "id": get_uuid(LayoutElements.LAYOUT),
            "content": "Dashboard vizualizing Eclipse well completion output.",
        },
        {
            "id": get_uuid(LayoutElements.ENSEMBLE_DROPDOWN),
            "content": "Select ensemble.",
        },
        {
            "id": get_uuid(LayoutElements.WELL_COMPLETIONS_COMPONENT),
            "content": (
                "Visualization of the well completions. "
                "Time slider for selecting which time steps to display. "
                "Different vizualisation and filtering alternatives are available "
                "in the upper right corner."
            ),
        },
    ]



import itertools
from enum import Enum
from typing import List

import numpy as np
import pandas
import plotly.express as px
import plotly.graph_objects as go

from webviz_subsurface._providers import EnsembleTableProvider


class _Columns(Enum):
    REALIZATION = "realization"
    VOLUME = "volume"
    CONTAINMENT = "containment"
    VOLUME_OUTSIDE = "volume_outside"


def _read_dataframe(
    table_provider: EnsembleTableProvider, realization: int
) -> pandas.DataFrame:
    return table_provider.get_column_data(
        ["date", "co2_inside", "co2_outside"], [realization]
    )


def _read_terminal_co2_volumes(
    table_provider: EnsembleTableProvider, realizations: List[int]
) -> pandas.DataFrame:
    records = []
    for real in realizations:
        df = _read_dataframe(table_provider, real)
        last = df.iloc[np.argmax(df["date"])]
        label = str(real)
        records += [
            (label, last["co2_inside"], "inside", 0.0),
            (label, last["co2_outside"], "outside", last["co2_outside"]),
        ]
    df = pandas.DataFrame.from_records(
        records,
        columns=[
            _Columns.REALIZATION.value,
            _Columns.VOLUME.value,
            _Columns.CONTAINMENT.value,
            _Columns.VOLUME_OUTSIDE.value,
        ],
    )
    df.sort_values(_Columns.VOLUME_OUTSIDE.value, inplace=True, ascending=True)
    return df


def _read_co2_volumes(
    table_provider: EnsembleTableProvider, realizations: List[int]
) -> pandas.DataFrame:
    return pandas.concat(
        [
            _read_dataframe(table_provider, real).assign(realization=real)
            for real in realizations
        ]
    )


def generate_co2_volume_figure(
    table_provider: EnsembleTableProvider,
    realizations: List[int],
) -> go.Figure:
    df = _read_terminal_co2_volumes(table_provider, realizations)
    fig = px.bar(
        df,
        y=_Columns.REALIZATION.value,
        x=_Columns.VOLUME.value,
        color=_Columns.CONTAINMENT.value,
        title="End-state CO2 containment",
        orientation="h",
        category_orders={_Columns.CONTAINMENT.value: ["outside", "inside"]},
        color_discrete_sequence=["#dd4300", "#006ddd"],
    )
    fig.layout.legend.title.text = ""
    fig.layout.legend.orientation = "h"
    fig.layout.legend.y = -0.3
    fig.layout.yaxis.title = "Realization"
    fig.layout.xaxis.exponentformat = "power"
    fig.layout.xaxis.title = "Mass [kg]"
    fig.layout.paper_bgcolor = "rgba(0,0,0,0)"
    fig.layout.margin.b = 10
    fig.layout.margin.t = 60
    fig.layout.margin.l = 10
    fig.layout.margin.r = 10
    return fig


def generate_co2_time_containment_figure(
    table_provider: EnsembleTableProvider,
    realizations: List[int],
) -> go.Figure:
    df = _read_co2_volumes(table_provider, realizations)
    df.sort_values(by="date", inplace=True)
    df["date"] = df["date"].astype(str)
    dates = df["date"].str[:4] + "-" + df["date"].str[4:6] + "-" + df["date"].str[6:]
    df["dt"] = dates
    df["co2_total"] = df["co2_inside"] + df["co2_outside"]
    fig = go.Figure()
    colors = px.colors.qualitative.Plotly
    # Generate dummy scatters for legend entries
    outside_args = dict(line_dash="dot", legendgroup="Outside", name="Outside")
    total_args = dict(legendgroup="Total", name="Total")
    dummy_args = dict(mode="lines", marker_color="black", hoverinfo="none")
    fig.add_scatter(y=[0.0], **dummy_args, **outside_args)
    fig.add_scatter(y=[0.0], **dummy_args, **total_args)
    for rlz, color in zip(realizations, itertools.cycle(colors)):
        sub_df = df[df["realization"] == rlz]
        common_args = dict(
            x=sub_df["dt"],
            hovertemplate="%{x}: %{y}<br>Realization: %{meta[0]}",
            meta=[rlz],
            marker_color=color,
            showlegend=False,
        )
        fig.add_scatter(y=sub_df["co2_outside"], **outside_args, **common_args)
        fig.add_scatter(y=sub_df["co2_total"], **total_args, **common_args)
    fig.layout.legend.orientation = "h"
    fig.layout.legend.title.text = ""
    fig.layout.legend.yanchor = "bottom"
    fig.layout.legend.y = 1.02
    fig.layout.legend.xanchor = "right"
    fig.layout.legend.x = 1
    fig.layout.xaxis.title = "Time (date)"
    fig.layout.yaxis.title = "Mass [kg]"
    fig.layout.paper_bgcolor = "rgba(0,0,0,0)"
    fig.layout.margin.b = 10
    fig.layout.margin.t = 60
    fig.layout.margin.l = 10
    fig.layout.margin.r = 10
    fig.layout.yaxis.range = (0, 1.05 * df["co2_total"].max())
    return fig



from typing import Callable

import webviz_core_components as wcc
from webviz_config import WebvizConfigTheme

from ..models import ParametersModel, SimulationTimeSeriesModel
from .parameter_qc_view import parameter_qc_view
from .parameter_response_view import parameter_response_view


def main_view(
    get_uuid: Callable,
    vectormodel: SimulationTimeSeriesModel,
    parametermodel: ParametersModel,
    theme: WebvizConfigTheme,
) -> wcc.Tabs:
    tabs = [
        wcc.Tab(
            label="Parameter distributions",
            value="tab-1",
            children=parameter_qc_view(
                get_uuid=get_uuid, parametermodel=parametermodel
            ),
        )
    ]

    if vectormodel is not None and parametermodel.mc_ensembles:
        tabs.append(
            wcc.Tab(
                label="Parameters impact on simulation profiles",
                value="tab-2",
                children=parameter_response_view(
                    get_uuid=get_uuid,
                    parametermodel=parametermodel,
                    vectormodel=vectormodel,
                    theme=theme,
                ),
            )
        )

    return wcc.Tabs(
        value="tab-1" if vectormodel is None else "tab-2",
        style={"width": "100%"},
        children=tabs,
    )



from typing import Callable

from dash import ALL, Input, Output, State, callback, callback_context, no_update
from dash.exceptions import PreventUpdate

from ..utils.utils import update_relevant_components


def layout_controllers(get_uuid: Callable) -> None:
    @callback(
        Output(get_uuid("page-selected"), "data"),
        Output(get_uuid("voldist-page-selected"), "data"),
        Input({"id": get_uuid("selections"), "button": ALL}, "n_clicks"),
        Input(get_uuid("tabs"), "value"),
        State({"id": get_uuid("selections"), "button": ALL}, "id"),
        State(get_uuid("voldist-page-selected"), "data"),
    )
    def _selected_page_controllers(
        _apply_click: int,
        tab_selected: str,
        button_ids: list,
        previous_page: dict,
    ) -> tuple:

        ctx = callback_context.triggered[0]
        initial_pages = {"voldist": "custom", "tornado": "torn_multi"}

        # handle initial callback
        if ctx["prop_id"] == ".":
            page_selected = (
                tab_selected
                if not tab_selected in initial_pages
                else initial_pages[tab_selected]
            )
            previous_page = initial_pages

        elif "tabs" in ctx["prop_id"]:
            page_selected = (
                tab_selected
                if not tab_selected in initial_pages
                else previous_page[tab_selected]
            )
            previous_page = no_update

        else:
            for button_id in button_ids:
                if button_id["button"] in ctx["prop_id"]:
                    page_selected = previous_page[tab_selected] = button_id["button"]

        return page_selected, previous_page

    @callback(
        Output({"id": get_uuid("selections"), "button": ALL}, "style"),
        Input(get_uuid("page-selected"), "data"),
        State(get_uuid("tabs"), "value"),
        State({"id": get_uuid("selections"), "button": ALL}, "id"),
    )
    def _update_button_style(
        page_selected: str, tab_selected: str, button_ids: list
    ) -> list:

        if tab_selected not in ["voldist", "tornado"]:
            raise PreventUpdate

        button_styles = {
            button["button"]: {"background-color": "#E8E8E8"} for button in button_ids
        }
        button_styles[page_selected] = {"background-color": "#7393B3", "color": "#fff"}

        return update_relevant_components(
            id_list=button_ids,
            update_info=[
                {
                    "new_value": style,
                    "conditions": {"button": button},
                }
                for button, style in button_styles.items()
            ],
        )

    @callback(
        Output({"id": get_uuid("main-voldist"), "page": ALL}, "style"),
        Input(get_uuid("page-selected"), "data"),
        State({"id": get_uuid("main-voldist"), "page": ALL}, "id"),
        State(get_uuid("tabs"), "value"),
    )
    def _main_voldist_display(
        page_selected: str,
        main_layout_ids: list,
        tab_selected: str,
    ) -> list:

        if tab_selected != "voldist":
            raise PreventUpdate

        voldist_layout = []
        for page_id in main_layout_ids:
            if page_id["page"] == page_selected:
                voldist_layout.append({"display": "block"})
            else:
                voldist_layout.append({"display": "none"})
        return voldist_layout

    @callback(
        Output({"id": get_uuid("main-tornado"), "page": ALL}, "style"),
        Input(get_uuid("page-selected"), "data"),
        State({"id": get_uuid("main-tornado"), "page": ALL}, "id"),
        State(get_uuid("tabs"), "value"),
    )
    def _main_tornado_display(
        page_selected: str,
        main_layout_ids: list,
        tab_selected: str,
    ) -> list:

        if tab_selected != "tornado":
            raise PreventUpdate

        main_layout = []
        for page_id in main_layout_ids:
            if page_id["page"] == page_selected:
                main_layout.append({"display": "block"})
            else:
                main_layout.append({"display": "none"})
        return main_layout



from typing import Dict, List, Tuple

import webviz_core_components as wcc
from dash import Input, Output, State, callback, dash_table, dcc, html
from dash.development.base_component import Component
from dash.exceptions import PreventUpdate
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

from .._types import DeltaEnsemble
from .._utils.delta_ensemble_utils import create_delta_ensemble_names


def _create_delta_ensemble_table_column_data(
    column_name: str, ensemble_names: List[str]
) -> List[Dict[str, str]]:
    return [{column_name: elm} for elm in ensemble_names]


class EnsemblesSettings(SettingsGroupABC):
    class Ids(StrEnum):
        ENSEMBLES_DROPDOWN = "ensembles-dropdown"
        DELTA_ENSEMBLE = "delta-ensemble"
        DELTA_ENSEMBLE_A_DROPDOWN = "delta-ensemble-a-dropdown"
        DELTA_ENSEMBLE_B_DROPDOWN = "delta-ensemble-b-dropdown"
        DELTA_ENSEMBLE_CREATE_BUTTON = "delta-ensemble-create-button"
        CREATED_DELTA_ENSEMBLES_STORE = "created-delta-ensembles-store"
        CREATED_DELTA_ENSEMBLE_NAMES_TABLE = "create-delta-ensemble-names-table"
        CREATED_DELTA_ENSEMBLE_NAMES_TABLE_COLUMN = (
            "create-delta-ensemble-names-table-column"
        )

    def __init__(
        self, ensembles_names: List[str], input_provider_set: EnsembleSummaryProviderSet
    ) -> None:
        super().__init__("Ensembles")
        self._ensembles = ensembles_names
        self._input_provider_set = input_provider_set

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Selected ensembles",
                id=self.register_component_unique_id(
                    EnsemblesSettings.Ids.ENSEMBLES_DROPDOWN
                ),
                clearable=False,
                multi=True,
                options=[
                    {"label": ensemble, "value": ensemble}
                    for ensemble in self._ensembles
                ],
                value=None if len(self._ensembles) <= 0 else [self._ensembles[0]],
            ),
            wcc.Selectors(
                label="Delta Ensembles",
                id=self.register_component_unique_id(
                    EnsemblesSettings.Ids.DELTA_ENSEMBLE
                ),
                children=self._delta_ensemble_creator_layout(),
            ),
        ]

    def _delta_ensemble_creator_layout(self) -> html.Div:
        return html.Div(
            children=[
                wcc.Dropdown(
                    label="Ensemble A",
                    id=self.register_component_unique_id(
                        EnsemblesSettings.Ids.DELTA_ENSEMBLE_A_DROPDOWN
                    ),
                    clearable=False,
                    options=[{"label": i, "value": i} for i in self._ensembles],
                    value=self._ensembles[0],
                    style={"min-width": "60px"},
                ),
                wcc.Dropdown(
                    label="Ensemble B",
                    id=self.register_component_unique_id(
                        EnsemblesSettings.Ids.DELTA_ENSEMBLE_B_DROPDOWN
                    ),
                    clearable=False,
                    options=[{"label": i, "value": i} for i in self._ensembles],
                    value=self._ensembles[-1],
                    style={"min-width": "60px"},
                ),
                html.Button(
                    "Create",
                    id=self.register_component_unique_id(
                        EnsemblesSettings.Ids.DELTA_ENSEMBLE_CREATE_BUTTON
                    ),
                    n_clicks=0,
                    style={
                        "margin-top": "5px",
                        "margin-bottom": "5px",
                        "min-width": "20px",
                    },
                ),
                self._delta_ensemble_table_layout(),
                dcc.Store(
                    id=self.register_component_unique_id(
                        EnsemblesSettings.Ids.CREATED_DELTA_ENSEMBLES_STORE
                    ),
                    data=[],
                ),  # TODO: Add predefined deltas?
            ]
        )

    def _delta_ensemble_table_layout(self) -> dash_table.DataTable:
        return dash_table.DataTable(
            id=self.register_component_unique_id(
                EnsemblesSettings.Ids.CREATED_DELTA_ENSEMBLE_NAMES_TABLE
            ),
            columns=(
                [
                    {
                        "id": self.register_component_unique_id(
                            EnsemblesSettings.Ids.CREATED_DELTA_ENSEMBLE_NAMES_TABLE_COLUMN
                        ),
                        "name": "Created Delta (A-B)",
                    }
                ]
            ),
            data=[],
            fixed_rows={"headers": True},
            style_as_list_view=True,
            style_cell={"textAlign": "left"},
            style_header={"fontWeight": "bold"},
            style_table={
                "maxHeight": "150px",
                "overflowY": "auto",
            },
            editable=False,
        )

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(
                    EnsemblesSettings.Ids.CREATED_DELTA_ENSEMBLES_STORE
                ).to_string(),
                "data",
            ),
            Output(
                self.component_unique_id(
                    EnsemblesSettings.Ids.CREATED_DELTA_ENSEMBLE_NAMES_TABLE
                ).to_string(),
                "data",
            ),
            Output(
                self.component_unique_id(
                    EnsemblesSettings.Ids.ENSEMBLES_DROPDOWN
                ).to_string(),
                "options",
            ),
            Input(
                self.component_unique_id(
                    EnsemblesSettings.Ids.DELTA_ENSEMBLE_CREATE_BUTTON
                ).to_string(),
                "n_clicks",
            ),
            State(
                self.component_unique_id(
                    EnsemblesSettings.Ids.CREATED_DELTA_ENSEMBLES_STORE
                ).to_string(),
                "data",
            ),
            State(
                self.component_unique_id(
                    EnsemblesSettings.Ids.DELTA_ENSEMBLE_A_DROPDOWN
                ).to_string(),
                "value",
            ),
            State(
                self.component_unique_id(
                    EnsemblesSettings.Ids.DELTA_ENSEMBLE_B_DROPDOWN
                ).to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _update_created_delta_ensembles_names(
            n_clicks: int,
            existing_delta_ensembles: List[DeltaEnsemble],
            ensemble_a: str,
            ensemble_b: str,
        ) -> Tuple[List[DeltaEnsemble], List[Dict[str, str]], List[Dict[str, str]]]:
            if n_clicks is None or n_clicks <= 0:
                raise PreventUpdate

            delta_ensemble = DeltaEnsemble(ensemble_a=ensemble_a, ensemble_b=ensemble_b)
            if delta_ensemble in existing_delta_ensembles:
                raise PreventUpdate

            new_delta_ensembles = existing_delta_ensembles
            new_delta_ensembles.append(delta_ensemble)

            # Create delta ensemble names
            new_delta_ensemble_names = create_delta_ensemble_names(new_delta_ensembles)

            table_data = _create_delta_ensemble_table_column_data(
                self.component_unique_id(
                    EnsemblesSettings.Ids.CREATED_DELTA_ENSEMBLE_NAMES_TABLE_COLUMN
                ).to_string(),
                new_delta_ensemble_names,
            )

            ensemble_options = [
                {"label": ensemble, "value": ensemble}
                for ensemble in self._input_provider_set.provider_names()
            ]
            for elm in new_delta_ensemble_names:
                ensemble_options.append({"label": elm, "value": elm})

            return (new_delta_ensembles, table_data, ensemble_options)



from typing import Callable, Optional

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from dash import Input, Output, State, callback, html
from dash.exceptions import PreventUpdate
from pandas.api.types import is_numeric_dtype

from webviz_subsurface._abbreviations.volume_terminology import (
    volume_description,
    volume_unit,
)
from webviz_subsurface._figures import create_figure
from webviz_subsurface._models import InplaceVolumesModel

from ..utils.table_and_figure_utils import (
    create_data_table,
    create_table_columns,
    fluid_annotation,
)
from ..utils.utils import move_to_end_of_list, to_ranges
from ..views.distribution_main_layout import (
    convergence_plot_layout,
    custom_plotting_layout,
    plots_per_zone_region_layout,
)


# pylint: disable=too-many-statements
def distribution_controllers(
    get_uuid: Callable, volumemodel: InplaceVolumesModel
) -> None:
    @callback(
        Output({"id": get_uuid("main-voldist"), "page": "custom"}, "children"),
        Input(get_uuid("selections"), "data"),
        State(get_uuid("page-selected"), "data"),
    )
    def _update_page_custom(selections: dict, page_selected: str) -> tuple:

        if page_selected != "custom":
            raise PreventUpdate

        selections = selections[page_selected]
        if not selections["update"]:
            raise PreventUpdate

        selected_data = [
            selections[x]
            for x in ["Subplots", "Color by", "X Response", "Y Response"]
            if selections[x] is not None
        ]
        groups = ["REAL"]
        parameters = []
        for item in selected_data:
            if item in volumemodel.selectors and item not in groups:
                groups.append(item)
            if item in volumemodel.parameters and item not in parameters:
                parameters.append(item)

        # for bo/bg the data should be grouped on fluid zone
        if any(x in selected_data for x in ["BO", "BG"]) and "FLUID_ZONE" not in groups:
            if "BO" in selected_data and "BG" in selected_data:
                return html.Div(
                    "Can't plot BO against BG", style={"margin-top": "40px"}
                )
            selections["filters"]["FLUID_ZONE"] = [
                "oil" if "BO" in selected_data else "gas"
            ]

        dframe = volumemodel.get_df(
            filters=selections["filters"], groups=groups, parameters=parameters
        )

        if dframe.empty:
            return html.Div(
                "No data left after filtering", style={"margin-top": "40px"}
            )

        df_for_figure = (
            dframe
            if not (selections["Plot type"] == "bar" and not "REAL" in selected_data)
            else dframe.groupby([x for x in groups if x != "REAL"]).mean().reset_index()
        )
        figure = (
            create_figure(
                plot_type=selections["Plot type"],
                data_frame=df_for_figure,
                x=selections["X Response"],
                y=selections["Y Response"],
                nbins=selections["hist_bins"],
                facet_col=selections["Subplots"],
                color=selections["Color by"],
                color_discrete_sequence=selections["Colorscale"],
                color_continuous_scale=selections["Colorscale"],
                barmode=selections["barmode"],
                boxmode=selections["barmode"],
                layout=dict(
                    title=dict(
                        text=(
                            f"{volume_description(selections['X Response'])}"
                            + (
                                f" [{volume_unit(selections['X Response'])}]"
                                if selections["X Response"]
                                in volumemodel.volume_columns
                                else ""
                            )
                        ),
                        x=0.5,
                        xref="paper",
                        font=dict(size=18),
                    ),
                ),
                yaxis=dict(showticklabels=True),
            )
            .add_annotation(fluid_annotation(selections))
            .update_xaxes({"matches": None} if not selections["X axis matches"] else {})
            .update_yaxes({"matches": None} if not selections["Y axis matches"] else {})
            .update_xaxes(
                {"type": "category", "tickangle": 45, "tickfont_size": 12}
                if selections["X Response"] in volumemodel.selectors
                else {}
            )
        )

        return custom_plotting_layout(
            figure=figure,
            tables=make_tables(
                dframe=dframe,
                responses=list({selections["X Response"], selections["Y Response"]}),
                groups=groups,
                volumemodel=volumemodel,
                page_selected=page_selected,
                selections=selections,
                table_type="Statistics table",
                view_height=37,
            )
            if selections["bottom_viz"] == "table"
            else None,
        )

    @callback(
        Output({"id": get_uuid("main-table"), "page": "table"}, "children"),
        Input(get_uuid("selections"), "data"),
        State(get_uuid("page-selected"), "data"),
    )
    def _update_page_tables(
        selections: dict,
        page_selected: str,
    ) -> list:

        if page_selected != "table":
            raise PreventUpdate

        selections = selections[page_selected]
        if not selections["update"]:
            raise PreventUpdate

        table_groups = (
            ["ENSEMBLE", "REAL"]
            if selections["Table type"] == "Statistics table"
            else ["ENSEMBLE"]
        )
        if selections["Group by"] is not None:
            table_groups.extend(
                [x for x in selections["Group by"] if x not in table_groups]
            )
        dframe = volumemodel.get_df(filters=selections["filters"], groups=table_groups)

        return make_tables(
            dframe=dframe,
            responses=selections["table_responses"],
            groups=selections["Group by"],
            view_height=85,
            table_type=selections["Table type"],
            volumemodel=volumemodel,
            page_selected=page_selected,
            selections=selections,
        )

    @callback(
        Output({"id": get_uuid("main-voldist"), "page": "per_zr"}, "children"),
        Input(get_uuid("selections"), "data"),
        State(get_uuid("page-selected"), "data"),
    )
    def _update_page_per_zr(selections: dict, page_selected: str) -> list:
        if page_selected != "per_zr":
            raise PreventUpdate

        selections = selections[page_selected]
        if not selections["update"]:
            raise PreventUpdate

        figs = []
        selectors = [
            x
            for x in ["ZONE", "REGION", "FACIES", "FIPNUM", "SET"]
            if x in volumemodel.selectors
        ]
        color = selections["Color by"] is not None
        for selector in selectors:
            groups = list({selector, selections["Color by"]}) if color else [selector]
            dframe = volumemodel.get_df(filters=selections["filters"], groups=groups)
            piefig = (
                (
                    create_figure(
                        plot_type="pie",
                        data_frame=dframe,
                        values=selections["X Response"],
                        names=selector,
                        color_discrete_sequence=selections["Colorscale"],
                        color=selector,
                    )
                    .update_traces(marker_line=dict(color="#000000", width=1))
                    .update_layout(margin=dict(l=10, b=10))
                )
                if not color
                else []
            )
            barfig = create_figure(
                plot_type="bar",
                data_frame=dframe,
                x=selector,
                y=selections["X Response"],
                title=f"{selections['X Response']} per {selector}",
                barmode="overlay" if selector == selections["Color by"] else "group",
                layout={"bargap": 0.05},
                color_discrete_sequence=selections["Colorscale"],
                color=selections["Color by"],
                text=selections["X Response"],
                xaxis=dict(type="category", tickangle=45, tickfont_size=17, title=None),
            ).update_traces(
                texttemplate=(
                    "%{text:.3s}"
                    if selections["X Response"] in volumemodel.volume_columns
                    else "%{text:.3g}"
                ),
                textposition="auto",
            )

            if selections["X Response"] not in volumemodel.hc_responses:
                barfig.add_annotation(fluid_annotation(selections))
            figs.append([piefig, barfig])
        return plots_per_zone_region_layout(figs)

    @callback(
        Output({"id": get_uuid("main-voldist"), "page": "conv"}, "children"),
        Input(get_uuid("selections"), "data"),
        State(get_uuid("page-selected"), "data"),
    )
    def _update_page_conv(selections: dict, page_selected: str) -> go.Figure:
        if page_selected != "conv":
            raise PreventUpdate

        selections = selections[page_selected]
        if not selections["update"]:
            raise PreventUpdate

        subplots = selections["Subplots"] if selections["Subplots"] is not None else []
        groups = ["REAL"]
        if subplots and subplots not in groups:
            groups.append(subplots)

        dframe = volumemodel.get_df(filters=selections["filters"], groups=groups)
        dframe = dframe.sort_values(by=["REAL"])

        if dframe.empty:
            return html.Div(
                "No data left after filtering", style={"margin-top": "40px"}
            )

        dfs = []
        df_groups = dframe.groupby(subplots) if subplots else [(None, dframe)]
        for _, df in df_groups:
            for calculation in ["mean", "p10", "p90"]:
                df_stat = df.reset_index(drop=True).copy()
                df_stat[selections["X Response"]] = (
                    (df_stat[selections["X Response"]].expanding().mean())
                    if calculation == "mean"
                    else df_stat[selections["X Response"]]
                    .expanding()
                    .quantile(0.1 if calculation == "p90" else 0.9)
                )
                df_stat["calculation"] = calculation
                df_stat["index"] = df_stat.index + 1
                dfs.append(df_stat)
        if dfs:
            dframe = pd.concat(dfs)

        title = (
            f"<b>Convergence plot of mean/p10/p90 for {selections['X Response']} </b>"
            "  -  shaded areas indicates failed/filtered out realizations"
        )

        figure = (
            create_figure(
                plot_type="line",
                data_frame=dframe,
                x="REAL",
                y=selections["X Response"],
                facet_col=selections["Subplots"],
                color="calculation",
                custom_data=["calculation", "index"],
                title=title,
                yaxis=dict(showticklabels=True),
            )
            .update_traces(
                hovertemplate=(
                    f"{selections['X Response']} %{{y}} <br>"
                    f"%{{customdata[0]}} for realizations {dframe['REAL'].min()}-%{{x}}<br>"
                    "Realization count: %{customdata[1]} <extra></extra>"
                ),
                line_width=3.5,
            )
            .update_traces(line_color="black", selector={"name": "mean"})
            .update_traces(
                line=dict(color="firebrick", dash="dash"), selector={"name": "p10"}
            )
            .update_traces(
                line=dict(color="royalblue", dash="dash"), selector={"name": "p90"}
            )
            .update_xaxes({"matches": None} if not selections["X axis matches"] else {})
            .update_yaxes({"matches": None} if not selections["Y axis matches"] else {})
        )
        if selections["X Response"] not in volumemodel.hc_responses:
            figure.add_annotation(fluid_annotation(selections))

        missing_reals = [
            x
            for x in range(dframe["REAL"].min(), dframe["REAL"].max())
            if x not in dframe["REAL"].unique()
        ]
        if missing_reals:
            for real_range in to_ranges(missing_reals):
                figure.add_vrect(
                    x0=real_range[0] - 0.5,
                    x1=real_range[1] + 0.5,
                    fillcolor="gainsboro",
                    layer="below",
                    opacity=0.4,
                    line_width=0,
                )

        return convergence_plot_layout(figure=figure)


# pylint: disable=too-many-locals
def make_tables(
    dframe: pd.DataFrame,
    responses: list,
    volumemodel: InplaceVolumesModel,
    selections: dict,
    table_type: str,
    view_height: float,
    page_selected: str,
    groups: Optional[list] = None,
) -> html.Div:

    groups = groups if groups is not None else []

    if table_type == "Statistics table":
        statcols = ["Mean", "Stddev", "P90", "P10", "Min", "Max"]
        groups = [x for x in groups if x != "REAL"]
        responses = [x for x in responses if x != "REAL" and x is not None]

        if not groups:
            df_groups = [("", dframe)]
        elif len(groups) == 1:
            df_groups = dframe.groupby(groups[0])
        else:
            df_groups = dframe.groupby(groups)
        data_properties = []
        data_volcols = []
        for response in responses:
            if not is_numeric_dtype(dframe[response]):
                continue
            for name, df in df_groups:
                values = df[response]
                data = {
                    "Mean": values.mean(),
                    "Stddev": values.std(),
                    "P10": np.nanpercentile(values, 90),
                    "P90": np.nanpercentile(values, 10),
                    "Min": values.min(),
                    "Max": values.max(),
                }
                if "FLUID_ZONE" not in groups:
                    data.update(
                        FLUID_ZONE=(" + ").join(selections["filters"]["FLUID_ZONE"])
                    )
                for idx, group in enumerate(groups):
                    data[group] = (
                        name
                        if name is not None or name is not isinstance(name, str)
                        else list(name)[idx]
                    )
                if response in volumemodel.volume_columns:
                    data["Response"] = response
                    data_volcols.append(data)
                else:
                    data["Property"] = response
                    data_properties.append(data)

        if data_volcols and data_properties:
            view_height = view_height / 2

        return [
            create_data_table(
                selectors=volumemodel.selectors,
                columns=create_table_columns(
                    columns=move_to_end_of_list(
                        "FLUID_ZONE", [col] + groups + statcols
                    ),
                    text_columns=[col] + groups,
                    use_si_format=statcols if col == "Response" else None,
                ),
                data=data,
                height=f"{view_height}vh",
                table_id={"table_id": f"{page_selected}-{col}"},
            )
            for col, data in zip(
                ["Response", "Property"], [data_volcols, data_properties]
            )
        ]

    # if table type Mean table
    groupby_real = (
        selections["Group by"] is not None and "REAL" in selections["Group by"]
    )
    if "REAL" in groups and not groupby_real:
        groups.remove("REAL")

    columns = responses + [x for x in groups if x not in responses]
    dframe = (
        dframe[columns].groupby(groups).mean().reset_index()
        if groups
        else dframe[responses].mean().to_frame().T
    )

    if "FLUID_ZONE" not in dframe:
        dframe["FLUID_ZONE"] = (" + ").join(selections["filters"]["FLUID_ZONE"])

    dframe = dframe[move_to_end_of_list("FLUID_ZONE", dframe.columns)]
    return [
        create_data_table(
            selectors=volumemodel.selectors,
            columns=create_table_columns(
                columns=dframe.columns, use_si_format=volumemodel.volume_columns
            ),
            data=dframe.iloc[::-1].to_dict("records"),
            height=f"{view_height}vh",
            table_id={"table_id": f"{page_selected}-meantable"},
        )
    ]
